#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import logging
import math
import numpy as np
from abc import abstractmethod
from collections import Counter
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.utils import check_random_state
from typing import TYPE_CHECKING, Literal

from .abstract_validation_utils import (
    AbstractValidationUtils,
    AbstractValidationUtilsClassification,
    AbstractValidationUtilsRegression,
)

if TYPE_CHECKING:
    from sklearn.base import BaseEstimator

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def caruana_weighted(
    predictions: list[np.ndarray],
    labels: np.ndarray,
    seed,
    n_iterations,
    loss_function,
):
    """Caruana's ensemble selection with replacement."""
    # -- Init Vars
    num_input_models_ = len(predictions)
    ensemble = []  # type: list[np.ndarray]
    trajectory = []  # contains iteration best
    val_loss_over_iterations_ = []  # contains overall best
    order = []
    rand = check_random_state(seed)
    weighted_ensemble_prediction = np.zeros(predictions[0].shape, dtype=np.float64)

    fant_ensemble_prediction = np.zeros(
        weighted_ensemble_prediction.shape,
        dtype=np.float64,
    )

    for i in range(n_iterations):
        logger.debug(f"Iteration {i}")

        ens_size = len(ensemble)
        if ens_size > 0:
            np.add(
                weighted_ensemble_prediction,
                ensemble[-1],
                out=weighted_ensemble_prediction,
            )

        # -- Process Iteration Solutions
        losses = np.zeros((len(predictions)), dtype=np.float64)

        for j, pred in enumerate(predictions):
            np.add(weighted_ensemble_prediction, pred, out=fant_ensemble_prediction)
            np.multiply(
                fant_ensemble_prediction,
                (1.0 / float(ens_size + 1)),
                out=fant_ensemble_prediction,
            )
            losses[j] = loss_function(labels, fant_ensemble_prediction)

        if i == 0:
            model_losses = losses.copy()

        # -- Eval Iteration results
        all_best = np.argwhere(losses == np.nanmin(losses)).flatten()
        best = rand.choice(all_best)  # break ties randomly
        ensemble_loss = losses[best]

        ensemble.append(predictions[best])
        trajectory.append(ensemble_loss)
        order.append(best)

        # Build Correct Validation loss list
        if not val_loss_over_iterations_:
            # Init
            val_loss_over_iterations_.append(ensemble_loss)
        elif val_loss_over_iterations_[-1] > ensemble_loss:
            # Improved
            val_loss_over_iterations_.append(ensemble_loss)
        else:
            # Not Improved
            val_loss_over_iterations_.append(val_loss_over_iterations_[-1])

        # -- Break for special cases
        #   - If we only have a pool of base models of size 1 (code found the single best model)
        #   - If we find a perfect ensemble/model, stop early
        if (len(predictions) == 1) or (ensemble_loss == 0):
            break

    indices_ = order
    trajectory_ = trajectory
    min_score = np.min(trajectory_)
    idx_best = trajectory_.index(min_score)
    indices_ = indices_[: idx_best + 1]
    n_iterations = idx_best + 1

    ensemble_members = Counter(indices_).most_common()
    weights = np.zeros((num_input_models_,), dtype=np.float64)

    for ensemble_member in ensemble_members:
        weight = float(ensemble_member[1]) / n_iterations
        weights[ensemble_member[0]] = weight

    if np.sum(weights) < 1:
        weights = weights / np.sum(weights)

    logger.info(f"Order of selections: {order}")
    logger.info(f"Val loss over iterations: {val_loss_over_iterations_}")
    logger.info(f"Model losses: {model_losses}")
    logger.info(f"Best weights: {weights}")

    return weights, val_loss_over_iterations_, model_losses


class GreedyWeightedEnsemble(AbstractValidationUtils):
    tune_temperature: bool = False

    def __init__(
        self,
        n_iterations: int,
        silo_top_n: int,
        model_family_per_estimator: list[str] | None = None,
    ):
        """GES implementation for sklearn-like models.

        Args:
            n_iterations: number of iterations for GES.
            silo_top_n: number of models for SiloTopN pruning.
            model_family_per_estimator: the model family for each estimator used for SiloTopN pruning.
        """
        self.n_iterations = n_iterations
        self.silo_top_n = silo_top_n
        self.model_family_per_estimator = model_family_per_estimator
        self._model_family_per_estimator = None  # in case of early stopping
        (
            self.ensemble,
            self.val_loss_over_iterations_,
            self.model_losses_,
            self.weights_,
        ) = (None, None, None, None)

    @abstractmethod
    def _build_ensemble(self, base_models, weights):
        pass

    def _tune_temperature_in_place(
        self,
        oof_proba_per_estimator: list[np.ndarray],
        y: np.ndarray,
    ) -> list[np.ndarray]:
        for model_i, (model_name, model) in enumerate(self._estimators):
            logger.info(f"Tuning temperature for model {model_name}")
            oof_proba_per_estimator[model_i] = self.tune_temperature_for_model_in_place(
                base_model=model,
                oof_predictions=oof_proba_per_estimator[model_i],
                y=y,
            )

    def get_weights(self, X, y):
        oof_proba = self.get_oof_per_estimator(X, y)
        self.model_family_per_estimator = (
            self.model_family_per_estimator
            if self.model_family_per_estimator is not None
            else ["X"] * len(self._estimators)
        )
        self._model_family_per_estimator = self.model_family_per_estimator[
            : len(self._estimators)
        ]

        if self._is_holdout:
            oof_proba = [oof[self._holdout_index_hit_list] for oof in oof_proba]
            y = y[self._holdout_index_hit_list]

        # If tune_temperature is True, we will tune the temperature for each model here before GES.
        if self.tune_temperature:
            self._tune_temperature_in_place(oof_proba_per_estimator=oof_proba, y=y)

        # Prune base models before GES
        data_for_pruning = []
        data_for_selection = {}
        for (bm_name, bm_model), bm_oof_proba, bm_family in zip(
            self._estimators,
            oof_proba,
            self._model_family_per_estimator,
            strict=True,
        ):
            bm_loss = self.loss_function(y, bm_oof_proba)
            data_for_selection[bm_name] = (bm_model, bm_oof_proba)
            data_for_pruning.append((bm_loss, bm_name, bm_family))
        bm_to_keep = _prune_to_silo_top_n(
            data_for_pruning,
            n=self.silo_top_n,
            maximize_metric=False,
        )

        tmp_estimators = []
        tmp_oof_proba = []
        for bm_name in bm_to_keep:
            bm_model, bm_oof_proba = data_for_selection[bm_name]
            tmp_estimators.append((bm_name, bm_model))
            tmp_oof_proba.append(bm_oof_proba)
        self._estimators = tmp_estimators
        oof_proba = tmp_oof_proba

        # Run GES
        (
            self.weights_,
            self.val_loss_over_iterations_,
            self.model_losses_,
        ) = caruana_weighted(
            oof_proba,
            y,
            self.seed + 62,
            n_iterations=self.n_iterations,
            loss_function=self.loss_function,
        )

        return self.weights_

    def fit(self, X, y):
        weights = self.get_weights(X, y)

        final_weights = []
        base_models = []
        for bm, weight in zip(self._estimators, weights, strict=True):
            if weight != 0:
                final_weights.append(weight)
                base_models.append(bm)

        self.ensemble = self._build_ensemble(base_models, final_weights)
        self.ensemble.fit(X, y)  # (re-)fit on full data
        return self

    def predict(self, X):
        return self.ensemble.predict(X)

    def predict_proba(self, X):
        return self.ensemble.predict_proba(X)


class GreedyWeightedEnsembleClassifier(
    GreedyWeightedEnsemble,
    AbstractValidationUtilsClassification,
):
    def __init__(
        self,
        estimators: list[tuple[str, BaseEstimator]],
        *,
        score_metric: str = "log_loss",
        time_limit: int | None = None,
        validation_method: Literal["cv", "holdout"] = "cv",
        holdout_fraction: float = 0.33,
        n_folds: int = 5,
        n_repeats: int = 1,
        seed: int = 42,
        n_iterations: int = 50,
        silo_top_n: int = 25,
        model_family_per_estimator: list[str] | None = None,
    ):
        """Post Hoc Ensemble Classifier Specialized for TabPFN models.

        Args:
            estimators: list of tuples of (name, estimator) containing the base models to ensemble
            score_metric: the metric string to use for scoring the ensemble models
            time_limit: the time limit for the validation process
            validation_method: the method to use for validation
            holdout_fraction: the fraction of the data to use for holdout validation
            n_folds: the number of folds to use for cross-validation
            n_repeats: the number of repeats to use for validation
            seed: the random seed used to control randomness
            n_iterations: the number of iterations to use for the greedy ensemble search
            silo_top_n: the total number of models used for ensemble selection.
                In other words, the number to prune to when using SiloTopN.
            model_family_per_estimator: the model family for each estimator used for SiloTopN pruning.
        """
        GreedyWeightedEnsemble.__init__(
            self,
            n_iterations=n_iterations,
            silo_top_n=silo_top_n,
            model_family_per_estimator=model_family_per_estimator,
        )
        AbstractValidationUtilsClassification.__init__(
            self,
            estimators=estimators,
            score_metric=score_metric,
            n_folds=n_folds,
            n_repeats=n_repeats,
            seed=seed,
            time_limit=time_limit,
            holdout_fraction=holdout_fraction,
            validation_method=validation_method,
        )

    def _build_ensemble(self, base_models, weights):
        return VotingClassifier(estimators=base_models, voting="soft", weights=weights)


class GreedyWeightedEnsembleRegressor(
    GreedyWeightedEnsemble,
    AbstractValidationUtilsRegression,
):
    def __init__(
        self,
        estimators: list[tuple[str, BaseEstimator]],
        *,
        score_metric: str = "rmse",
        validation_method: Literal["cv", "holdout"] = "cv",
        n_folds: int = 5,
        holdout_fraction: float = 0.33,
        n_repeats: int = 1,
        time_limit: int | None = None,
        seed: int = 42,
        n_iterations: int = 50,
        silo_top_n: int = 25,
        model_family_per_estimator: list[str] | None = None,
    ):
        """Post Hoc Ensemble Regressor Specialized for TabPFN models.

        Args:
            estimators: list of tuples of (name, estimator) containing the base models to ensemble
            score_metric: the metric string to use for scoring the ensemble models
            time_limit: the time limit for the validation process
            validation_method: the method to use for validation
            holdout_fraction: the fraction of the data to use for holdout validation
            n_folds: the number of folds to use for cross-validation
            n_repeats: the number of repeats to use for validation
            seed: the random seed used to control randomness
            n_iterations: the number of iterations to use for the greedy ensemble search
            silo_top_n: the total number of models used for ensemble selection.
                In other words, the number to prune to when using SiloTopN.
            model_family_per_estimator: the model family for each estimator used for SiloTopN pruning.
        """
        GreedyWeightedEnsemble.__init__(
            self,
            n_iterations=n_iterations,
            silo_top_n=silo_top_n,
            model_family_per_estimator=model_family_per_estimator,
        )
        AbstractValidationUtilsRegression.__init__(
            self,
            estimators=estimators,
            score_metric=score_metric,
            n_folds=n_folds,
            n_repeats=n_repeats,
            seed=seed,
            time_limit=time_limit,
            holdout_fraction=holdout_fraction,
            validation_method=validation_method,
        )

    def _build_ensemble(self, base_models, weights):
        return VotingRegressor(estimators=base_models, weights=weights)


def _prune_to_silo_top_n(
    base_models_data: list[tuple[float, str, str]],
    *,
    n: int,
    maximize_metric: bool,
) -> list[str]:
    """Prune to the top N models per silo based on validation score. A silo represents an algorithm family.

    Code Taken from Lennart's (unpublished) version of Assembled (https://github.com/ISG-Siegen/assembled)

    Args:
        base_models_data: list of tuples of (metric_value, model_name, algorithm_family) per base model
        n: number of models to keep
        maximize_metric: whether to maximize the metric or not
    """
    # -- Get silos
    # algorithm family (af) to list of base models of this family
    af_to_model = {}  # type: dict[str, list[tuple[float, str, str]]]
    for bm_metric_value, bm_name, bm_family in base_models_data:
        if bm_family not in af_to_model:
            af_to_model[bm_family] = []

        af_to_model[bm_family].append((bm_metric_value, bm_name, bm_family))

    # Get the minimal number of entire a silo can have (in case of equal distribution)
    min_silo_val = max(math.floor(n / len(af_to_model.keys())), 1)

    while sum(len(base_models) for base_models in af_to_model.values()) > n:
        # Find silos with too many values
        too_large_silos = [
            af
            for af, base_models in af_to_model.items()
            if len(base_models) > min_silo_val
        ]

        if not too_large_silos:
            break

        # For all silos with too many values, find and remove the model with the worst performance
        #   - This won't remove silos entirely, because silos with at least 1 element won't be too large
        #   - The first element of sorted() is the element with the worst performance (highest/lowest value)
        worst_model = sorted(
            [base_model for af in too_large_silos for base_model in af_to_model[af]],
            key=lambda x: x[0],  # sort by validation score/loss
            reverse=maximize_metric,  # determines if higher or lower is better
        )[
            -1
        ]  # select the worst model (first element)
        af_to_model[worst_model[-1]].remove(worst_model)

    if sum(len(base_models) for base_models in af_to_model.values()) > n:
        # In this case, we have more silos than top_n (cat_to_model.keys() > top_n)
        # Moreover, at this point, all silos will only have 1 element in it.
        # To resolve this, we can simply return the top_n models over these silos
        # (other fallbacks like random for more diversity would work as well, but we think top is best for now)
        sort_rest = sorted(
            [
                base_model
                for base_models in af_to_model.values()
                for base_model in base_models
            ],  # all models
            key=lambda x: x[0],
            reverse=maximize_metric,
        )
        bm_names_mv_to_keep = [
            (name, metric_val) for metric_val, name, _ in sort_rest[-n:]
        ]
    else:
        bm_names_mv_to_keep = [
            (name, metric_val)
            for vals in af_to_model.values()
            for metric_val, name, _ in vals
        ]

    # Sort to have similar order to top n (also can be beneficial for ensembling)
    return [
        x[0]
        for x in sorted(
            bm_names_mv_to_keep,
            key=lambda m: m[1],
            reverse=maximize_metric,
        )
    ]



"""TODO: Don't copy data but keep indices instead."""

#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import logging
import numpy as np
import time
import torch
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.ensemble._base import _partition_estimators
from sklearn.utils.validation import (
    check_is_fitted,
)

from .SklearnBasedDecisionTreeTabPFN import (
    DecisionTreeTabPFNClassifier,
    DecisionTreeTabPFNRegressor,
)

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger("RF-PFN")


def softmax_numpy(logits):
    exp_logits = np.exp(logits)  # Apply exponential to each logit
    sum_exp_logits = np.sum(
        exp_logits,
        axis=-1,
        keepdims=True,
    )  # Sum of exponentials across classes
    return exp_logits / sum_exp_logits  # Normalize to get probabilities


class RandomForestTabPFNBase:
    """Base Class for common functionalities."""

    def get_n_estimators(self, X):
        return self.n_estimators

    def set_categorical_features(self, categorical_features):
        """Sets categorical features
        :param categorical_features: Categorical features
        :return: None.
        """
        self.categorical_features = categorical_features

    def fit(self, X, y, sample_weight=None):
        """Fits RandomForestTabPFN
        :param X: Feature training data
        :param y: Label training data
        :param sample_weight: Weights of each sample
        :return: None.
        """
        self.estimator = self.init_base_estimator()
        self.estimator.set_categorical_features(self.categorical_features)

        self.X = X
        self.n_estimators = self.get_n_estimators(X)

        time.time()
        if self.max_depth == 0:
            self.tabpfn.fit(X, y)
        else:
            try:
                if torch.is_tensor(X):
                    X = X.numpy()
                if torch.is_tensor(y):
                    y = y.numpy()
                super().fit(X, y)
            except TypeError as e:
                print("Error in fit with data", X, y)
                raise e

        return self


class RandomForestTabPFNClassifier(RandomForestTabPFNBase, RandomForestClassifier):
    """RandomForestTabPFNClassifier."""

    task_type = "multiclass"

    def __init__(
        self,
        tabpfn=None,
        n_jobs=1,
        categorical_features=None,
        show_progress=False,
        verbose=0,
        adaptive_tree=True,
        fit_nodes=True,
        adaptive_tree_overwrite_metric="log_loss",
        adaptive_tree_test_size=0.2,
        adaptive_tree_min_train_samples=100,
        adaptive_tree_max_train_samples=5000,
        adaptive_tree_min_valid_samples_fraction_of_train=0.2,
        preprocess_X_once=False,
        max_predict_time=60,
        rf_average_logits=True,
        dt_average_logits=True,
        adaptive_tree_skip_class_missing=True,
        # Added to make cloneable.
        n_estimators=100,
        criterion="gini",
        max_depth=5,
        min_samples_split=1000,
        min_samples_leaf=5,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        random_state=None,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0.0,
        max_samples=None,
    ):
        """:param N_ensemble_configurations: Parameter of TabPFNClassifier
        :param n_estimators: Number of DT-TabPFN models
        :param min_samples_split: min_samples_split param of DTTabPFN
        :param max_features: sklearn DT param
        :param bootstrap: Whether use bootstrapping or not
        :param n_jobs: parallel processing of submodels n_jobs will be processed simultaniously
        :param random_state: random state of sklearn RF model
        """
        super().__init__(
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            verbose=verbose,
            n_jobs=n_jobs,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            bootstrap=bootstrap,
            oob_score=oob_score,
            random_state=random_state,
            warm_start=warm_start,
            class_weight=class_weight,
            ccp_alpha=ccp_alpha,
            max_samples=max_samples,
        )

        self.tabpfn = tabpfn

        self.categorical_features = categorical_features
        self.show_progress = show_progress
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.adaptive_tree = adaptive_tree
        self.fit_nodes = fit_nodes
        self.adaptive_tree_overwrite_metric = adaptive_tree_overwrite_metric
        self.adaptive_tree_test_size = adaptive_tree_test_size
        self.adaptive_tree_min_train_samples = adaptive_tree_min_train_samples
        self.adaptive_tree_max_train_samples = adaptive_tree_max_train_samples
        self.adaptive_tree_min_valid_samples_fraction_of_train = (
            adaptive_tree_min_valid_samples_fraction_of_train
        )
        self.preprocess_X_once = preprocess_X_once
        self.max_predict_time = max_predict_time
        self.rf_average_logits = rf_average_logits
        self.dt_average_logits = dt_average_logits
        self.adaptive_tree_skip_class_missing = adaptive_tree_skip_class_missing
        self.n_estimators = n_estimators

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        tags.estimator_type = "classifier"
        return tags

    def init_base_estimator(self):
        return DecisionTreeTabPFNClassifier(
            tabpfn=self.tabpfn,
            min_samples_split=self.min_samples_split,
            max_features=self.max_features,
            random_state=self.random_state,
            categorical_features=self.categorical_features,
            max_depth=self.max_depth,
            show_progress=self.show_progress,
            adaptive_tree=self.adaptive_tree,
            fit_nodes=self.fit_nodes,
            verbose=self.verbose,
            adaptive_tree_test_size=self.adaptive_tree_test_size,
            adaptive_tree_overwrite_metric=self.adaptive_tree_overwrite_metric,
            adaptive_tree_min_train_samples=self.adaptive_tree_min_train_samples,
            adaptive_tree_max_train_samples=self.adaptive_tree_max_train_samples,
            adaptive_tree_min_valid_samples_fraction_of_train=self.adaptive_tree_min_valid_samples_fraction_of_train,
            average_logits=self.dt_average_logits,
            adaptive_tree_skip_class_missing=self.adaptive_tree_skip_class_missing,
        )

    def predict(self, X):
        """Predict class for X.

        The predicted class of an input sample is a vote by the trees in
        the forest, weighted by their probability estimates. That is,
        the predicted class is the one with highest mean probability
        estimate across the trees.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns:
        -------
        y : ndarray of shape (n_samples,) or (n_samples, n_outputs)
            The predicted classes.
        """
        proba = self.predict_proba(X)

        if self.n_outputs_ == 1:
            return self.classes_.take(np.argmax(proba, axis=1), axis=0)

        n_samples = proba[0].shape[0]
        # all dtypes should be the same, so just take the first
        class_type = self.classes_[0].dtype
        predictions = np.empty((n_samples, self.n_outputs_), dtype=class_type)

        for k in range(self.n_outputs_):
            predictions[:, k] = self.classes_[k].take(
                np.argmax(proba[k], axis=1),
                axis=0,
            )

        return predictions

    def _final_proba(self, all_proba, evaluated_estimators):
        for proba in all_proba:
            proba /= evaluated_estimators

        if self.rf_average_logits:
            # convert all_proba logits for the multiclass output to probabilities per class
            all_proba = softmax_numpy(all_proba)

        if len(all_proba) == 1:
            return all_proba[0]
        return all_proba

    def predict_proba(self, X):
        """Predict class probabilities for X.

        The predicted class probabilities of an input sample are computed as
        the mean predicted class probabilities of the trees in the forest.
        The class probability of a single tree is the fraction of samples of
        the same class in a leaf.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns:
        -------
        p : ndarray of shape (n_samples, n_classes), or a list of such arrays
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        # Check data
        X = self._validate_X_predict(X)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        # avoid storing the output of every estimator by summing them here
        all_proba = [
            np.zeros((X.shape[0], j), dtype=np.float64)
            for j in np.atleast_1d(self.n_classes_)
        ]

        start_time = time.time()
        evaluated_estimators = 0
        for e in self.estimators_:
            _accumulate_prediction(
                e.predict_proba,
                X,
                all_proba,
                accumulate_logits=self.rf_average_logits,
            )

            time_elapsed = time.time() - start_time
            evaluated_estimators += 1
            if time_elapsed > self.max_predict_time and self.max_predict_time > 0:
                break

        return self._final_proba(all_proba, evaluated_estimators)


class RandomForestTabPFNRegressor(RandomForestTabPFNBase, RandomForestRegressor):
    """RandomForestTabPFNRegressor."""

    task_type = "regression"

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        tags.estimator_type = "regressor"
        return tags

    def __init__(
        self,
        tabpfn=None,
        n_jobs=1,
        categorical_features=None,
        show_progress=False,
        verbose=0,
        adaptive_tree=True,
        fit_nodes=True,
        adaptive_tree_overwrite_metric="rmse",
        adaptive_tree_test_size=0.2,
        adaptive_tree_min_train_samples=100,
        adaptive_tree_max_train_samples=5000,
        adaptive_tree_min_valid_samples_fraction_of_train=0.2,
        preprocess_X_once=False,
        max_predict_time=-1,
        rf_average_logits=False,
        # Added to make cloneable.
        n_estimators=16,
        criterion="friedman_mse",
        max_depth=5,
        min_samples_split=300,
        min_samples_leaf=5,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        random_state=None,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None,
    ):
        """:param N_ensemble_configurations: Parameter of TabPFNClassifier
        :param n_estimators: Number of DT-TabPFN models
        :param min_samples_split: min_samples_split param of DTTabPFN
        :param max_features: sklearn DT param
        :param bootstrap: Whether use bootstrapping or not
        :param n_jobs: parallel processing of submodels n_jobs will be processed simultaniously
        :param random_state: random state of sklearn RF model
        """
        super().__init__(
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            ccp_alpha=ccp_alpha,
            max_samples=max_samples,
        )

        self.tabpfn = tabpfn

        self.categorical_features = categorical_features
        self.show_progress = show_progress
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.adaptive_tree = adaptive_tree
        self.fit_nodes = fit_nodes
        self.adaptive_tree_overwrite_metric = adaptive_tree_overwrite_metric
        self.adaptive_tree_test_size = adaptive_tree_test_size
        self.adaptive_tree_min_train_samples = adaptive_tree_min_train_samples
        self.adaptive_tree_max_train_samples = adaptive_tree_max_train_samples
        self.adaptive_tree_min_valid_samples_fraction_of_train = (
            adaptive_tree_min_valid_samples_fraction_of_train
        )
        self.preprocess_X_once = preprocess_X_once
        self.max_predict_time = max_predict_time
        self.rf_average_logits = rf_average_logits

    def init_base_estimator(self):
        return DecisionTreeTabPFNRegressor(
            tabpfn=self.tabpfn,
            min_samples_split=self.min_samples_split,
            max_features=self.max_features,
            random_state=self.random_state,
            categorical_features=self.categorical_features,
            max_depth=self.max_depth,
            show_progress=self.show_progress,
            adaptive_tree=self.adaptive_tree,
            fit_nodes=self.fit_nodes,
            verbose=self.verbose,
            adaptive_tree_test_size=self.adaptive_tree_test_size,
            adaptive_tree_overwrite_metric=self.adaptive_tree_overwrite_metric,
            adaptive_tree_min_train_samples=self.adaptive_tree_min_train_samples,
            adaptive_tree_max_train_samples=self.adaptive_tree_max_train_samples,
            adaptive_tree_min_valid_samples_fraction_of_train=self.adaptive_tree_min_valid_samples_fraction_of_train,
        )

    def predict(self, X):
        """Predict regression target for X.

        The predicted regression target of an input sample is computed as the
        mean predicted regression targets of the trees in the forest.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns:
        -------
        y : ndarray of shape (n_samples,) or (n_samples, n_outputs)
            The predicted values.
        """
        check_is_fitted(self)
        # Check data
        X = self._validate_X_predict(X)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        # avoid storing the output of every estimator by summing them here
        y_hat = (
            np.zeros((X.shape[0], self.n_outputs_), dtype=np.float64)
            if self.n_outputs_ > 1
            else np.zeros(X.shape[0], dtype=np.float64)
        )

        start_time = time.time()
        evaluated_estimators = 0

        for e in self.estimators_:
            _accumulate_prediction(e.predict, X, [y_hat])
            time_elapsed = time.time() - start_time
            evaluated_estimators += 1
            if time_elapsed > self.max_predict_time and self.max_predict_time > 0:
                break

        y_hat /= evaluated_estimators

        return y_hat


def _accumulate_prediction(predict, X, out, accumulate_logits=False):
    """This is a utility function for joblib's Parallel.

    It can't go locally in ForestClassifier or ForestRegressor, because joblib
    complains that it cannot pickle it when placed there.
    """
    prediction = predict(X, check_input=False)

    if accumulate_logits:
        # convert multiclass probabilities to logits
        prediction = np.log(prediction + 1e-10)  # Add small value to avoid log(0)

    if len(out) == 1:
        out[0] += prediction
    else:
        for i in range(len(out)):
            out[i] += prediction[i]



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0
from typing import Optional, Union

import numpy as np
import pandas as pd
import tabpfn


def get_tabpfn_explainer(
    model: Union[tabpfn.TabPFNRegressor, tabpfn.TabPFNClassifier],
    data: Union[pd.DataFrame, np.ndarray],
    labels: Union[pd.DataFrame, np.ndarray],
    index: str = "k-SII",
    max_order: int = 2,
    class_index: Optional[int] = None,
    **kwargs
):
    """Get a TabPFNExplainer from shapiq.

    This function returns the TabPFN explainer from the shapiq[1]_ library. The explainer uses
    a remove-and-recontextualize paradigm of model explanation[2]_[3]_ to explain the predictions
    of a TabPFN model. See ``shapiq.TabPFNExplainer`` documentation for more information regarding
    the explainer object.

    Args:
        model (tabpfn.TabPFNRegressor or tabpfn.TabPFNClassifier): The TabPFN model to explain.

        data (pd.DataFrame or np.ndarray): The background data to use for the explainer.

        labels (pd.DataFrame or np.ndarray): The labels for the background data.

        index: The index to use for the explanation. See shapiq documentation for more information
            and an up-to-date list of available indices. Defaults to "k-SII" and "SV" (Shapley
            Values like SHAP) with ``max_order=1``.

        max_order (int): The maximum order of interactions to consider. Defaults to 2.

        class_index (int, optional): The class index of the model to explain. If not provided, the
            class index will be set to 1 per default for classification models. This argument is
            ignored for regression models. Defaults to None.

        **kwargs: Additional keyword arguments to pass to the explainer.

    Returns:
        shapiq.TabPFNExplainer: The TabPFN explainer.

    References:
        .. [1] shapiq repository: https://github.com/mmschlk/shapiq
        .. [2] Muschalik, M., Baniecki, H., Fumagalli, F., Kolpaczki, P., Hammer, B., Hüllermeier, E. (2024). shapiq: Shapley Interactions for Machine Learning. In: The Thirty-eight Conference on Neural Information Processing Systems Datasets and Benchmarks Track. url: https://openreview.net/forum?id=knxGmi6SJi
        .. [3] Rundel, D., Kobialka, J., von Crailsheim, C., Feurer, M., Nagler, T., Rügamer, D. (2024). Interpretable Machine Learning for TabPFN. In: Longo, L., Lapuschkin, S., Seifert, C. (eds) Explainable Artificial Intelligence. xAI 2024. Communications in Computer and Information Science, vol 2154. Springer, Cham. https://doi.org/10.1007/978-3-031-63797-1_23

    """
    import shapiq

    # make data to array if it is a pandas DataFrame
    if isinstance(data, pd.DataFrame):
        data = data.values

    # make labels to array if it is a pandas Series
    if isinstance(labels, pd.Series) or isinstance(labels, pd.DataFrame):
        labels = labels.values

    return shapiq.TabPFNExplainer(
        model=model,
        data=data,
        labels=labels,
        index=index,
        max_order=max_order,
        class_index=class_index,
        **kwargs,
    )


def get_tabpfn_imputation_explainer(
    model: Union[tabpfn.TabPFNRegressor, tabpfn.TabPFNClassifier],
    data: Union[pd.DataFrame, np.ndarray],
    index: str = "k-SII",
    max_order: int = 2,
    imputer: str = "marginal",
    class_index: Optional[int] = None,
    **kwargs
):
    """Gets a TabularExplainer from shapiq with using imputation.

    This function returns the TabularExplainer from the shapiq[1]_[2]_ library. The explainer uses an
    imputation-based paradigm of feature removal for the explanations similar to SHAP[3]_. See
    ``shapiq.TabularExplainer`` documentation for more information regarding the explainer object.

    Args:
        model (tabpfn.TabPFNRegressor or tabpfn.TabPFNClassifier): The TabPFN model to explain.

        data (pd.DataFrame or np.ndarray): The background data to use for the explainer.

        index: The index to use for the explanation. See shapiq documentation for more information
            and an up-to-date list of available indices. Defaults to "k-SII" and "SV" (Shapley
            Values like SHAP) with ``max_order=1``.

        max_order (int): The maximum order of interactions to consider. Defaults to 2.

        imputer: The imputation method to use. See ``shapiq.TabularExplainer`` documentation for
            more information and an up-to-date list of available imputation methods.

        class_index (int, optional): The class index of the model to explain. If not provided, the
            class index will be set to 1 per default for classification models. This argument is
            ignored for regression models. Defaults to None.

        **kwargs: Additional keyword arguments to pass to the explainer.

    Returns:
        shapiq.TabularExplainer: The TabularExplainer.

    References:
        .. [1] shapiq repository: https://github.com/mmschlk/shapiq
        .. [2] Muschalik, M., Baniecki, H., Fumagalli, F., Kolpaczki, P., Hammer, B., Hüllermeier, E. (2024). shapiq: Shapley Interactions for Machine Learning. In: The Thirty-eight Conference on Neural Information Processing Systems Datasets and Benchmarks Track. url: https://openreview.net/forum?id=knxGmi6SJi
        .. [3] Lundberg, S. M., & Lee, S. I. (2017). A Unified Approach to Interpreting Model Predictions. Advances in Neural Information Processing Systems 30 (pp. 4765--4774).

    """
    import shapiq

    # make data to array if it is a pandas DataFrame
    if isinstance(data, pd.DataFrame):
        data = data.values

    return shapiq.TabularExplainer(
        model=model,
        data=data,
        index=index,
        max_order=max_order,
        imputer=imputer,
        class_index=class_index,
        **kwargs,
    )



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import numpy as np
import random
import torch
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils import check_random_state
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted
from typing import Literal

from .pfn_phe import (
    AutoPostHocEnsemblePredictor,
    TaskType,
)

MAX_INT = int(np.iinfo(np.int32).max)


class AutoTabPFNClassifier(ClassifierMixin, BaseEstimator):
    """Automatic Post Hoc Ensemble Classifier for TabPFN models.

    Parameters
    ----------
        max_time : int | None, default=None
            The maximum time to spend on fitting the post hoc ensemble.
        preset: {"default", "custom_hps", "avoid_overfitting"}, default="default"
            The preset to use for the post hoc ensemble.
        ges_scoring_string : str, default="roc"
            The scoring string to use for the greedy ensemble search.
            Allowed values are: {"accuracy", "roc" / "auroc", "f1", "log_loss"}.
        device : {"cpu", "cuda"}, default="cuda"
            The device to use for training and prediction.
        random_state : int, RandomState instance or None, default=None
            Controls both the randomness base models and the post hoc ensembling method.
        categorical_feature_indices: list[int] or None, default=None
            The indices of the categorical features in the input data. Can also be passed to `fit()`.
        ignore_pretraining_limits: bool, default=False
            Whether to ignore the pretraining limits of the TabPFN base models.
        phe_init_args : dict | None, default=None
            The initialization arguments for the post hoc ensemble predictor.
            See post_hoc_ensembles.pfn_phe.AutoPostHocEnsemblePredictor for more options and all details.

    Attributes:
    ----------
        predictor_ : AutoPostHocEnsemblePredictor
            The predictor interface used to make predictions, see post_hoc_ensembles.pfn_phe.AutoPostHocEnsemblePredictor for more.
        phe_init_args_ : dict
            The optional initialization arguments used for the post hoc ensemble predictor.
    """

    predictor_: AutoPostHocEnsemblePredictor
    phe_init_args_: dict
    n_features_in_: int

    def __init__(
        self,
        max_time: int | None = 30,
        preset: Literal["default", "custom_hps", "avoid_overfitting"] = "default",
        ges_scoring_string: str = "roc",
        device: Literal["cpu", "cuda"] = "cpu",
        random_state: int | None | np.random.RandomState = None,
        categorical_feature_indices: list[int] | None = None,
        ignore_pretraining_limits: bool = False,
        phe_init_args: dict | None = None,
    ):
        self.max_time = max_time
        self.preset = preset
        self.ges_scoring_string = ges_scoring_string
        self.device = device
        self.random_state = random_state
        self.phe_init_args = phe_init_args
        self.categorical_feature_indices = categorical_feature_indices
        self.ignore_pretraining_limits = ignore_pretraining_limits

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        tags.estimator_type = "classifier"
        return tags

    def fit(self, X, y, categorical_feature_indices: list[int] | None = None):
        if categorical_feature_indices is not None:
            self.categorical_feature_indices = categorical_feature_indices

        self.phe_init_args_ = {} if self.phe_init_args is None else self.phe_init_args
        rnd = check_random_state(self.random_state)

        # Torch reproducibility bomb
        torch.manual_seed(rnd.randint(0, MAX_INT))
        random.seed(rnd.randint(0, MAX_INT))
        np.random.seed(rnd.randint(0, MAX_INT))  # noqa: NPY002

        task_type = (
            TaskType.MULTICLASS if len(unique_labels(y)) > 2 else TaskType.BINARY
        )
        self.predictor_ = AutoPostHocEnsemblePredictor(
            preset=self.preset,
            task_type=task_type,
            max_time=self.max_time,
            ges_scoring_string=self.ges_scoring_string,
            device=self.device,
            bm_random_state=rnd.randint(0, MAX_INT),
            ges_random_state=rnd.randint(0, MAX_INT),
            ignore_pretraining_limits=self.ignore_pretraining_limits,
            **self.phe_init_args_,
        )

        self.predictor_.fit(
            X,
            y,
            categorical_feature_indices=self.categorical_feature_indices,
        )

        # -- Sklearn required values
        self.classes_ = self.predictor_._label_encoder.classes_
        self.n_features_in_ = self.predictor_.n_features_in_

        return self

    def predict(self, X):
        check_is_fitted(self)
        return self.predictor_.predict(X)

    def predict_proba(self, X):
        check_is_fitted(self)
        return self.predictor_.predict_proba(X)

    def _more_tags(self):
        return {"allow_nan ": True}


class AutoTabPFNRegressor(RegressorMixin, BaseEstimator):
    """Automatic Post Hoc Ensemble Regressor for TabPFN models.

    Parameters
    ----------
        max_time : int | None, default=None
            The maximum time to spend on fitting the post hoc ensemble.
        preset: {"default", "custom_hps", "avoid_overfitting"}, default="default"
            The preset to use for the post hoc ensemble.
        ges_scoring_string : str, default="mse"
            The scoring string to use for the greedy ensemble search.
            Allowed values are: {"rmse", "mse", "mae"}.
        device : {"cpu", "cuda"}, default="cuda"
            The device to use for training and prediction.
        random_state : int, RandomState instance or None, default=None
            Controls both the randomness base models and the post hoc ensembling method.
        categorical_feature_indices: list[int] or None, default=None
            The indices of the categorical features in the input data. Can also be passed to `fit()`.
        ignore_pretraining_limits: bool, default=False
            Whether to ignore the pretraining limits of the TabPFN base models.
        phe_init_args : dict | None, default=None
            The initialization arguments for the post hoc ensemble predictor.
            See post_hoc_ensembles.pfn_phe.AutoPostHocEnsemblePredictor for more options and all details.

    Attributes:
    ----------
        predictor_ : AutoPostHocEnsemblePredictor
            The predictor interface used to make predictions, see post_hoc_ensembles.pfn_phe.AutoPostHocEnsemblePredictor for more.
        phe_init_args_ : dict
            The optional initialization arguments used for the post hoc ensemble predictor.
    """

    predictor_: AutoPostHocEnsemblePredictor
    phe_init_args_: dict
    n_features_in_: int

    def __init__(
        self,
        max_time: int | None = 30,
        preset: Literal["default", "custom_hps", "avoid_overfitting"] = "default",
        ges_scoring_string: str = "mse",
        device: Literal["cpu", "cuda"] = "cpu",
        random_state: int | None | np.random.RandomState = None,
        categorical_feature_indices: list[int] | None = None,
        ignore_pretraining_limits: bool = False,
        phe_init_args: dict | None = None,
    ):
        self.max_time = max_time
        self.preset = preset
        self.ges_scoring_string = ges_scoring_string
        self.device = device
        self.random_state = random_state
        self.phe_init_args = phe_init_args
        self.categorical_feature_indices = categorical_feature_indices
        self.ignore_pretraining_limits = ignore_pretraining_limits

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        tags.estimator_type = "regressor"
        return tags

    def fit(self, X, y, categorical_feature_indices: list[int] | None = None):
        if categorical_feature_indices is not None:
            self.categorical_feature_indices = categorical_feature_indices

        self.phe_init_args_ = {} if self.phe_init_args is None else self.phe_init_args
        rnd = check_random_state(self.random_state)

        # Torch reproducibility bomb
        torch.manual_seed(rnd.randint(0, MAX_INT))
        random.seed(rnd.randint(0, MAX_INT))
        np.random.seed(rnd.randint(0, MAX_INT))  # noqa: NPY002

        self.predictor_ = AutoPostHocEnsemblePredictor(
            preset=self.preset,
            task_type=TaskType.REGRESSION,
            max_time=self.max_time,
            ges_scoring_string=self.ges_scoring_string,
            device=self.device,
            bm_random_state=rnd.randint(0, MAX_INT),
            ges_random_state=rnd.randint(0, MAX_INT),
            ignore_pretraining_limits=self.ignore_pretraining_limits,
            **self.phe_init_args_,
        )

        self.predictor_.fit(
            X,
            y,
            categorical_feature_indices=self.categorical_feature_indices,
        )

        # -- Sklearn required values
        self.n_features_in_ = self.predictor_.n_features_in_

        return self

    def predict(self, X):
        check_is_fitted(self)
        return self.predictor_.predict(X)

    def _more_tags(self):
        return {"allow_nan ": True}


if __name__ == "__main__":
    import os

    from sklearn.utils.estimator_checks import check_estimator

    os.environ["SK_COMPATIBLE_PRECISION"] = "True"
    raise_on_error = True
    nan_test = 9

    # Precision issues do not allow for such deterministic behavior as expected, thus retrying certain tests to show it can work.
    clf_non_deterministic_for_reasons = [
        31,
        30,
    ]
    reg_non_deterministic_for_reasons = [
        27,
        28,
    ]

    for est, non_deterministic in [
        (AutoTabPFNClassifier(device="cuda"), clf_non_deterministic_for_reasons),
        (AutoTabPFNRegressor(device="cuda"), reg_non_deterministic_for_reasons),
    ]:
        print("Run for", est)
        lst = []
        for i, x in enumerate(check_estimator(est, generate_only=True)):
            if (i == nan_test) and ("allow_nan" in x[0]._get_tags()):
                # sklearn test does not check for the tag!
                continue

            n_tests = 5
            while n_tests:
                try:
                    x[1](x[0])
                except Exception as e:
                    print("Error in test:", i)
                    print(x)
                    if i in non_deterministic:
                        print("Non deterministic test, retrying")
                        n_tests -= 1
                        continue
                    if raise_on_error:
                        raise e
                    lst.append((i, x, e))
                break

    print("\n\n\n\n\n\nFailed tests")
    print(len(lst))
    print(lst)
    print([x[0] for x in lst])



from .SklearnBasedDecisionTreeTabPFN import (
    DecisionTreeTabPFNClassifier,
    DecisionTreeTabPFNRegressor,
)
from .SklearnBasedRandomForestTabPFN import (
    RandomForestTabPFNClassifier,
    RandomForestTabPFNRegressor,
)
from .configs import TabPFNRFConfig

__all__ = [
    "DecisionTreeTabPFNClassifier",
    "DecisionTreeTabPFNRegressor",
    "RandomForestTabPFNClassifier",
    "RandomForestTabPFNRegressor",
    "TabPFNRFConfig",
]



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import itertools
import logging
import numpy as np
import warnings
from enum import Enum
from sklearn.base import BaseEstimator
from sklearn.utils import check_array, check_X_y
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder
from typing import Literal

from tabpfn_extensions import TabPFNClassifier, TabPFNRegressor
from .greedy_weighted_ensemble import (
    GreedyWeightedEnsembleClassifier,
    GreedyWeightedEnsembleRegressor,
)
from ..rf_pfn.SklearnBasedRandomForestTabPFN import (
    RandomForestTabPFNClassifier,
    RandomForestTabPFNRegressor,
)

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger("AutoPostHocEnsemble")


class TaskType(str, Enum):
    BINARY = "binary_classification"
    MULTICLASS = "multiclass_classification"
    REGRESSION = "regression"


class PresetType(str, Enum):
    DEFAULT = "default"
    AVOID_OVERFITTING = "avoid_overfitting"


class DeviceType(str, Enum):
    CPU = "cpu"
    CUDA = "cuda"


class TabPFNBaseModelSource(str, Enum):
    CUSTOM = "custom"
    RANDOM_PORTFOLIO = "random_portfolio"


class AutoPostHocEnsemblePredictor(BaseEstimator):
    """A wrapper to effectively performing post hoc ensemble with TabPFN models."""

    # Task Config
    preset: PresetType
    task_type: TaskType
    device: DeviceType
    bm_random_state: int
    ges_random_state: int

    # Validation Data Config
    n_folds: int | None
    n_repeats: int
    validation_method: Literal["holdout", "cv"]
    holdout_fraction: float

    # GES config
    ges_max_time: int
    ges_n_iterations: int
    ges_scoring_string: str

    # Base Model config
    max_models: int | None
    tabpfn_base_model_source: TabPFNBaseModelSource
    custom_tabpfn_models: list[tuple[str, BaseEstimator]] | None
    additional_sklearn_models: list[tuple[str, BaseEstimator]] | None

    # Attributes
    _ens_model: (
        GreedyWeightedEnsembleClassifier | GreedyWeightedEnsembleRegressor
    )  # Ensemble model to use.
    _estimators: list[tuple[str, BaseEstimator]]  # Base models used in the ensemble.
    _cat_encoder: ColumnTransformer  # Encoder to use for categorical features before passing it to the base models.
    _label_encoder: LabelEncoder  # Label encoder to use for label features before passing it to the base models.

    def __init__(
        self,
        *,
        preset: PresetType,
        max_time: int | None,
        task_type: TaskType,
        ges_scoring_string: str,
        device: DeviceType,
        bm_random_state: int,
        ges_random_state: int,
        # -- Custom hyperparameters
        # - How to get base models
        tabpfn_base_model_source: TabPFNBaseModelSource = TabPFNBaseModelSource.RANDOM_PORTFOLIO,
        custom_tabpfn_models: list[tuple[str, BaseEstimator]] | None = None,
        additional_sklearn_models: list[tuple[str, BaseEstimator]] | None = None,
        # - How to use base models
        max_models: int | None = None,
        validation_method: Literal["holdout", "cv"] = "holdout",
        n_repeats: int = 80,
        n_folds: int | None = None,
        holdout_fraction: float = 0.33,
        ges_n_iterations: int = 25,
        ignore_pretraining_limits: bool = False,
    ) -> None:
        """Builds a PostHocEnsembleConfig with default values for the given parameters.

        Args:
            preset: The preset to use, which prompts changes to the method. Note, these
                - "default" use default hyperparameter. Does not override user-supplied parameters.
                - "avoid_overfitting" uses a preset optimized for avoiding overfitting (needs more training time!).
                    This will override certain parameters no-matter the user input.
            max_time: The maximum time to run the ensemble builder (only respected with an overhead).
            task_type: The task type of the dataset. Either "regression", "binary_classification", "multiclass_classification".
            ges_scoring_string: The scoring string to use for the greedy ensemble search.
                Supported values for classification: "accuracy", "roc" / "auroc", "f1", "log_loss".
                Supported values for regression: "rmse", "mse", "mae".
            device: The device to run the models on. Either "cpu" or "cuda".
            bm_random_state: The random state to use for the base models.
            ges_random_state: The random state to use for the greedy ensemble search.
            tabpfn_base_model_source: Where the TabPFN base models are sourced from.
                - `custom` uses models supplied via the custom_tabpfn_models parameter.
                - `random_portfolio` uses models from a random portfolio.
                - `zeroshot_portfolio` uses models from meta-learned portfolio, as defined in (the default) `zero_shot_portfolio_config`.
            custom_tabpfn_models: Custom TabPFN models to use for the post hoc ensemble.
                Only used if tabpfn_base_model_source is set to "custom".
            additional_sklearn_models: Additional sklearn models to use. If None, no additional models are used.
            max_models: The maximum number of base models to use. If None, all models are used.
            validation_method: The validation method to use. Either "holdout" or "cv".
            n_repeats: The number of repeats to use for holdout or cross-validation.
            n_folds: The number of folds to use for cross-validation. Pas None in the holdout setting.
            holdout_fraction: The fraction of the data to use for holdout validation.
            ges_n_iterations: The number of iterations to use for the greedy ensemble search.
            ignore_pretraining_limits: Whether to ignore the pretraining limits of the TabPFN models.
        """
        # Task Type and User Input
        self.preset = preset
        self.max_time = max_time
        self.task_type = task_type
        self.ges_scoring_string = ges_scoring_string
        self.device = device
        self.bm_random_state = bm_random_state
        self.ges_random_state = ges_random_state
        self.ignore_pretraining_limits = ignore_pretraining_limits

        # Model Source
        self.tabpfn_base_model_source = tabpfn_base_model_source
        self.custom_tabpfn_models = custom_tabpfn_models
        self.additional_sklearn_models = additional_sklearn_models

        # Model Usage
        self.max_models = max_models
        self.validation_method = validation_method
        self.n_repeats = n_repeats
        self.n_folds = n_folds
        self.holdout_fraction = holdout_fraction
        self.ges_n_iterations = ges_n_iterations

        # -- Input Validation
        if self.task_type not in TaskType:
            raise ValueError(f"Only supports {list(TaskType)}! Got: {task_type}")

        # -- Handle time
        if max_time in [None, -1, 0]:
            warnings.warn(
                "No max_time or max_time=-1 or max_time=0 given... ensembling two models and evaluate both with holdout validation!",
                stacklevel=2,
            )
            warnings.warn(
                "Set a time limit to train and ensemble more models!",
                stacklevel=2,
            )
            ges_max_time = float("inf")
            self.max_models = 2
            self.n_repeats = 1
            self.holdout_fraction = 0.2
            self.validation_method = "holdout"
        elif isinstance(max_time, int) and max_time > 0:
            ges_max_time = max_time
        else:
            raise NotImplementedError(
                f"max_time {max_time} not supported for TabPFN yet!",
            )
        self.ges_max_time = ges_max_time

        # -- Handle presets
        if self.preset == "default":
            logger.info("Using `default` preset for Post Hoc Ensemble.", stacklevel=2)
        elif self.preset == "avoid_overfitting":
            logger.info(
                "Using preset to avoid overfitting parameters. Enforcing cross-validation with 8 folds.",
                stacklevel=2,
            )
            self.validation_method = "cv"
            self.n_folds = 8  # AutoGluon value
        else:
            raise ValueError(f"preset {preset} not supported yet!")

        if (
            self.tabpfn_base_model_source == TabPFNBaseModelSource.CUSTOM
            and self.custom_tabpfn_models is None
        ):
            raise ValueError(
                "Custom models are specified as the source of TabPFN but not custom TabPFN models are supplied!",
            )

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        categorical_feature_indices: list[int] | None = None,
    ) -> AutoPostHocEnsemblePredictor:
        """Fits the post hoc ensemble on the given data.

        Args:
            X: The input data to fit the ensemble on.
            y: The target values to fit the ensemble on.
            categorical_feature_indices: The indices of the categorical features in the data.
                If None, no categorical features are assumed to be present.
        """
        if categorical_feature_indices is None:
            logger.info(
                "No categorical_feature_indices given. Assuming no categorical features.",
                stacklevel=2,
            )
            categorical_feature_indices = []
        elif all(isinstance(x, (int, np.integer)) for x in categorical_feature_indices):
            logger.info(
                f"Using categorical_feature_indices: {categorical_feature_indices}",
                stacklevel=2,
            )
        else:
            raise ValueError(
                f"Invalid categorical_feature_indices: {categorical_feature_indices}",
            )

        # -- Ensure dtype encoding.
        X, y = check_X_y(
            X,
            y,
            force_all_finite="allow-nan",
            dtype=object,
            accept_sparse=False,
        )
        self.n_features_in_ = X.shape[1]

        # -- Data Cleaning to be usable by base models.
        self._cat_encoder = ColumnTransformer(
            transformers=[
                (
                    "categorical",
                    OrdinalEncoder(
                        handle_unknown="use_encoded_value",
                        unknown_value=-1,
                    ),
                    categorical_feature_indices,
                ),
            ],
            remainder="passthrough",
            sparse_threshold=0,
        )
        X = self._cat_encoder.fit_transform(X)

        if self.task_type in [TaskType.BINARY, TaskType.MULTICLASS]:
            self._label_encoder = LabelEncoder()
            y = self._label_encoder.fit_transform(y)

        # -- sanity check, no trust.
        X, y = check_X_y(
            X,
            y,
            force_all_finite="allow-nan",
            dtype="numeric",
            accept_sparse=False,
        )

        if self.task_type in [TaskType.BINARY, TaskType.MULTICLASS]:
            self._ens_model: GreedyWeightedEnsembleClassifier = (
                GreedyWeightedEnsembleClassifier
            )
            self._pred_func = "predict_proba"
        elif self.task_type == TaskType.REGRESSION:
            self._ens_model: GreedyWeightedEnsembleRegressor = (
                GreedyWeightedEnsembleRegressor
            )
            self._pred_func = "predict"
        else:
            raise ValueError(f"Unknown task type {self.task_type}!")
        logger.info(f"Using task type: {self.task_type}", stacklevel=2)

        if self.validation_method == "cv" and self.task_type in [
            "multiclass_classification",
            "binary_classification",
        ]:
            # n_folds sanity check
            _, counts = np.unique(y, return_counts=True)
            second_largest: int = int(np.partition(counts.flatten(), -2)[-2])
            if second_largest < self.n_folds:
                warnings.warn(
                    "Detected that `n_folds` is hardcoded but dataset has less minimal non-prune-able classes. "
                    "We adjust n_folds to avoid cross-validation bugs in that case to the minimal class count. "
                    f"Old value: {self.n_folds}, new value: {second_largest}.",
                    stacklevel=2,
                )
                self.n_folds = second_largest

        self._estimators, model_family_per_estimator = self._collect_base_models(
            categorical_feature_indices=categorical_feature_indices,
        )

        self._ens_model = self._ens_model(
            estimators=self._estimators,
            seed=self.ges_random_state,
            n_repeats=self.n_repeats,
            n_folds=self.n_folds,
            n_iterations=self.ges_n_iterations,
            score_metric=self.ges_scoring_string,
            time_limit=self.ges_max_time,
            validation_method=self.validation_method,
            holdout_fraction=self.holdout_fraction,
            model_family_per_estimator=model_family_per_estimator,
        )

        self._ens_model.fit(X, y)

        return self

    def _collect_base_models(
        self,
        categorical_feature_indices: list[int],
    ) -> tuple[list[tuple[str, object]], list[str]]:
        assert isinstance(categorical_feature_indices, list)

        if self.tabpfn_base_model_source == TabPFNBaseModelSource.CUSTOM:
            logger.info(
                "Obtaining TabPFN models from user-defined models.",
                stacklevel=2,
            )
            bm_list = self.custom_tabpfn_models
        elif self.tabpfn_base_model_source == TabPFNBaseModelSource.RANDOM_PORTFOLIO:
            logger.info(
                "Obtaining TabPFN models from a random portfolio.",
                stacklevel=2,
            )
            bm_list = _get_base_models_from_random_search(
                task_type=self.task_type,
                device=self.device,
                random_state=self.bm_random_state,
                categorical_indices=categorical_feature_indices,
                ignore_pretraining_limits=self.ignore_pretraining_limits,
            )
        else:
            raise ValueError(
                f"Unknown base model source {self.tabpfn_base_model_source}!",
            )

        if self.additional_sklearn_models is not None:
            bm_list = _interleave_lists(bm_list, self.additional_sklearn_models)

        if self.max_models is not None:
            bm_list = bm_list[: self.max_models]

        logger.info(
            f"Using {len(bm_list)} base models: {[x[0] for x in bm_list]}",
            stacklevel=2,
        )

        model_family_per_bm = []
        for _, bm in bm_list:
            if isinstance(
                bm,
                (RandomForestTabPFNClassifier, RandomForestTabPFNRegressor),
            ):
                mf = "rf-pfn" + str(bm.tabpfn.model_path)
            elif isinstance(bm, (TabPFNClassifier, TabPFNRegressor)):
                mf = "pfn" + str(bm.model_path)
            elif isinstance(bm, BaseEstimator):
                mf = "sk-non-pfn"
            else:
                mf = "NAN"
            model_family_per_bm.append(mf)

        return bm_list, model_family_per_bm

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predicts the target values for the given data."""
        X = check_array(X, force_all_finite="allow-nan", dtype=object)
        X = check_array(
            self._cat_encoder.transform(X),
            force_all_finite="allow-nan",
            dtype="numeric",
        )
        if self.task_type == "regression":
            return self._ens_model.predict(X)

        return self._label_encoder.inverse_transform(self._ens_model.predict(X))

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        """Predicts the target values for the given data."""
        X = check_array(X, force_all_finite="allow-nan", dtype=object)
        X = check_array(
            self._cat_encoder.transform(X),
            force_all_finite="allow-nan",
            dtype="numeric",
        )
        return self._ens_model.predict_proba(X)


def _get_base_models_from_random_search(
    *,
    task_type: str,
    device: str,
    random_state: int,
    categorical_indices: list[int],
    random_portfolio_size: int = 100,
    start_with_default_pfn: bool = True,
    ignore_pretraining_limits: bool = False,
) -> list[tuple[str, object]]:
    # TODO: switch to config space to not depend on hyperopt
    from hyperopt.pyll import stochastic

    from ..hpo.search_space import get_param_grid_hyperopt

    _task_type = (
        "multiclass"
        if task_type in [TaskType.MULTICLASS, TaskType.BINARY]
        else task_type
    )
    rng = np.random.default_rng(4224)
    bm_model_rng = np.random.default_rng(random_state + 4224)
    search_space = get_param_grid_hyperopt(_task_type)
    model_base = TabPFNClassifier if _task_type == "multiclass" else TabPFNRegressor

    bm_list = []
    first_non_rf_pfn = None
    for i in range(random_portfolio_size):
        model_seed = int(bm_model_rng.integers(0, np.iinfo(np.int32).max))

        param = (
            {}
            if (start_with_default_pfn and i == 0)
            else stochastic.sample(search_space, rng=rng)
        )

        inference_config = {
            _k: v
            for k, v in param.items()
            if k.startswith("inference_config/") and (_k := k.split("/")[-1])
        }
        if inference_config:
            param["inference_config"] = inference_config
        for k in list(param.keys()):
            if k.startswith("inference_config/"):
                del param[k]

        param["device"] = device
        param["random_state"] = model_seed
        param["categorical_features_indices"] = categorical_indices
        param["ignore_pretraining_limits"] = ignore_pretraining_limits
        n_ensemble_repeats = param.pop("n_ensemble_repeats", None)
        model_is_rf_pfn = param.pop("model_type", "no") == "dt_pfn"
        if model_is_rf_pfn:
            param["n_estimators"] = 1
            rf_model_base = (
                RandomForestTabPFNClassifier
                if _task_type == "multiclass"
                else RandomForestTabPFNRegressor
            )
            bm = rf_model_base(
                tabpfn=model_base(**param),
                categorical_features=categorical_indices,
            )
        else:
            if n_ensemble_repeats is not None:
                param["n_estimators"] = n_ensemble_repeats
            bm = model_base(**param)

        model_indicator = "rf_pfn" if model_is_rf_pfn else "tabpfn"
        model_prefix = "default" if (start_with_default_pfn and i == 0) else "random"
        bm_list.append((f"{model_prefix}_{model_indicator}_model_{i}", bm))

    if first_non_rf_pfn is not None:
        bm_list = [first_non_rf_pfn, *bm_list]

    return bm_list


def _interleave_lists(list1, list2):
    return [
        x
        for x in itertools.chain(*itertools.zip_longest(list1, list2))
        if x is not None
    ]



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

from hyperopt import hp
from pathlib import Path


def enumerate_preprocess_transforms():
    transforms = []

    names_list = [
        ["safepower"],
        ["quantile_uni_coarse"],
        ["quantile_norm_coarse"],
        #["norm_and_kdi"],
        ["quantile_uni"],
        ["none"],
        ["robust"],
        ["safepower", "quantile_uni"],
        ["none", "safepower"],
    ]

    # try:
    #     from kditransform import KDITransformer
    #
    #     names_list += [
    #         ["kdi_uni"],
    #         ["kdi_alpha_0.3"],
    #         ["kdi_alpha_3.0"],
    #         ["kdi", "quantile_uni"],
    #     ]
    # except:
    #     pass

    for names in names_list:
        for categorical_name in [
            "numeric",
            "ordinal_very_common_categories_shuffled",
            "onehot",
            "none",
        ]:
            for append_original in [True, False]:
                for subsample_features in [-1, 0.99, 0.95, 0.9]:
                    for global_transformer_name in [None, "svd"]:
                        transforms += [
                            [
                                {
                                    "name": name,
                                    "global_transformer_name": global_transformer_name,
                                    "subsample_features": subsample_features,
                                    "categorical_name": categorical_name,
                                    "append_original": append_original,
                                }
                                for name in names
                            ],
                        ]
    return transforms


def get_param_grid_hyperopt(task_type: str) -> dict:
    search_space = {
        # Custom HPs
        "model_type": hp.choice("model_type", ["single", "dt_pfn"]),
        "n_ensemble_repeats": hp.choice("n_ensemble_repeats", [4]),
        # -- Model HPs
        "average_before_softmax": hp.choice("average_before_softmax", [True, False]),
        "softmax_temperature": hp.choice(
            "softmax_temperature",
            [
                0.75,
                0.8,
                0.9,
                0.95,
                1.0,
            ],
        ),
        # Inference config
        "inference_config/FINGERPRINT_FEATURE": hp.choice(
            "FINGERPRINT_FEATURE",
            [True, False],
        ),
        "inference_config/PREPROCESS_TRANSFORMS": hp.choice(
            "PREPROCESS_TRANSFORMS",
            enumerate_preprocess_transforms(),
        ),
        "inference_config/POLYNOMIAL_FEATURES": hp.choice(
            "POLYNOMIAL_FEATURES", ["no", 50]
        ),
        "inference_config/OUTLIER_REMOVAL_STD": hp.choice(
            "OUTLIER_REMOVAL_STD", [None, 7.0, 9.0, 12.0]
        ),
        "inference_config/SUBSAMPLE_SAMPLES": hp.choice(
            "SUBSAMPLE_SAMPLES", [0.99, None]
        ),
    }

    local_dir = (Path(__file__).parent / "hpo_models").resolve()

    if task_type == "multiclass":
        model_paths = [
            str(local_dir / "tabpfn-v2-classifier.ckpt"),
            str(local_dir / "tabpfn-v2-classifier-od3j1g5m.ckpt"),
            str(local_dir / "tabpfn-v2-classifier-gn2p4bpt.ckpt"),
            str(local_dir / "tabpfn-v2-classifier-znskzxi4.ckpt"),
            str(local_dir / "tabpfn-v2-classifier-llderlii.ckpt"),
            str(local_dir / "tabpfn-v2-classifier-vutqq28w.ckpt"),
        ]
    elif task_type == "regression":
        model_paths = [
            str(local_dir / "tabpfn-v2-regressor-09gpqh39.ckpt"),
            str(local_dir / "tabpfn-v2-regressor.ckpt"),
            str(local_dir / "tabpfn-v2-regressor-2noar4o2.ckpt"),
            str(local_dir / "tabpfn-v2-regressor-wyl4o83o.ckpt"),
            str(local_dir / "tabpfn-v2-regressor-5wof9ojf.ckpt"),
        ]
        search_space["inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS"] = hp.choice(
            "REGRESSION_Y_PREPROCESS_TRANSFORMS",
            [
                (None,),
                (None, "safepower"),
                ("safepower",),
                # ("kdi_alpha_0.3",),
                # ("kdi_alpha_1.0",),
                # ("kdi_alpha_1.5",),
                # ("kdi_alpha_0.6",),
                # ("kdi_alpha_3.0",),
                ("quantile_uni",),
            ],
        )
    else:
        raise ValueError(f"Unknown task type {task_type} for the search space!")

    search_space["model_path"] = hp.choice("model_path", model_paths)

    return search_space



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import logging
import numpy as np
import torch
import warnings
from enum import Enum
from hyperopt import fmin, tpe, STATUS_OK, Trials
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, check_array
from sklearn.compose import ColumnTransformer
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted, check_X_y
from typing import Literal, Optional

from tabpfn_extensions.hpo.search_space import get_param_grid_hyperopt
from tabpfn_extensions import TabPFNClassifier, TabPFNRegressor

logger = logging.getLogger(__name__)


class MetricType(str, Enum):
    ACCURACY = "accuracy"
    ROC_AUC = "roc_auc"
    F1 = "f1"
    RMSE = "rmse"
    MSE = "mse"
    MAE = "mae"


class TunedTabPFNBase(BaseEstimator):
    """Base class for tuned TabPFN models with proper categorical handling."""

    def __init__(
        self,
        n_trials: int = 50,
        n_validation_size: float = 0.2,
        metric: MetricType = MetricType.ACCURACY,
        device: str = "cpu",
        random_state: Optional[int] = None,
        categorical_feature_indices: Optional[list[int]] = None,
        verbose: bool = True,
    ):
        self.n_trials = n_trials
        self.n_validation_size = n_validation_size
        self.metric = MetricType(metric)  # Validate metric type
        self.device = device
        self.random_state = random_state
        self.categorical_feature_indices = categorical_feature_indices or []
        self.verbose = verbose

    def _setup_data_encoders(
        self, X: np.ndarray, categorical_feature_indices: Optional[list[int]] = None
    ):
        """Set up categorical and label encoders."""
        if categorical_feature_indices is not None:
            self.categorical_feature_indices = categorical_feature_indices

        if not self.categorical_feature_indices:
            logger.info(
                "No categorical features specified. Using all features as numeric."
            )
            self.categorical_feature_indices = []

        # Create categorical feature encoder
        self._cat_encoder = ColumnTransformer(
            transformers=[
                (
                    "categorical",
                    OrdinalEncoder(
                        handle_unknown="use_encoded_value",
                        unknown_value=-1,
                    ),
                    self.categorical_feature_indices,
                ),
            ],
            remainder="passthrough",
            sparse_threshold=0,
        )

    def _optimize(self, X: np.ndarray, y: np.ndarray, task_type: str):
        """Optimize hyperparameters using hyperopt with proper data handling."""
        rng = check_random_state(self.random_state)

        # Set random seeds for reproducibility
        torch.manual_seed(rng.randint(0, 2**31 - 1))
        np.random.seed(rng.randint(0, 2**31 - 1))

        # Fit transformers
        X = self._cat_encoder.fit_transform(X)
        if hasattr(self, "_label_encoder"):
            y = self._label_encoder.transform(y)

        # Split data for validation
        X_train, X_val, y_train, y_val = train_test_split(
            X,
            y,
            test_size=self.n_validation_size,
            random_state=rng.randint(0, 2**31 - 1),
            stratify=y if task_type == "multiclass" else None,
        )

        search_space = get_param_grid_hyperopt(
            "multiclass" if task_type in ["binary", "multiclass"] else "regression"
        )

        def objective(params):
            inference_config = {
                k.split("/")[-1]: v
                for k, v in params.items()
                if k.startswith("inference_config/") and (_k := k.split("/")[-1])
            }

            # Clean up params
            model_params = {
                k: v for k, v in params.items() if not k.startswith("inference_config/")
            }
            model_params["inference_config"] = inference_config
            model_params["device"] = self.device
            model_params["random_state"] = rng.randint(0, 2**31 - 1)

            # Handle special parameters
            n_ensemble_repeats = model_params.pop("n_ensemble_repeats", None)
            if n_ensemble_repeats is not None:
                model_params["n_estimators"] = n_ensemble_repeats

            # Handle model type - skip RF-PFN for now
            model_type = model_params.pop("model_type", "single")
            if model_type == "dt_pfn":
                return {"loss": float("inf"), "status": STATUS_OK}

            try:
                # Create and fit model
                if task_type in ["binary", "multiclass"]:
                    model = TabPFNClassifier(**model_params)
                else:
                    model = TabPFNRegressor(**model_params)

                model.fit(X_train, y_train)

                # Evaluate based on metric
                if task_type in ["binary", "multiclass"]:
                    if self.metric == MetricType.ACCURACY:
                        score = model.score(X_val, y_val)
                    elif self.metric in [MetricType.ROC_AUC]:
                        from sklearn.metrics import roc_auc_score

                        y_pred = model.predict_proba(X_val)
                        score = roc_auc_score(y_val, y_pred, multi_class="ovr")
                    elif self.metric == MetricType.F1:
                        from sklearn.metrics import f1_score

                        y_pred = model.predict(X_val)
                        score = f1_score(y_val, y_pred, average="weighted")
                else:
                    y_pred = model.predict(X_val)
                    from sklearn.metrics import mean_squared_error, mean_absolute_error

                    if self.metric in [MetricType.RMSE, MetricType.MSE]:
                        score = -mean_squared_error(
                            y_val, y_pred, squared=self.metric == MetricType.MSE
                        )
                    elif self.metric == MetricType.MAE:
                        score = -mean_absolute_error(y_val, y_pred)

                return {"loss": -score, "status": STATUS_OK, "model": model}

            except Exception as e:
                if self.verbose:
                    logger.warning(f"Trial failed with error: {str(e)}")
                return {"loss": float("inf"), "status": STATUS_OK}

        trials = Trials()
        best = fmin(
            fn=objective,
            space=search_space,
            algo=tpe.suggest,
            max_evals=self.n_trials,
            trials=trials,
            verbose=self.verbose,
        )

        # Store results
        self.best_params_ = best
        self.best_score_ = -min(trials.losses())
        self.best_model_ = trials.best_trial["result"].get("model")

        # If all trials failed, create default model
        if self.best_model_ is None:
            warnings.warn(
                "All optimization trials failed. Creating default model.",
                stacklevel=2,
            )
            if task_type in ["binary", "multiclass"]:
                self.best_model_ = TabPFNClassifier(
                    device=self.device,
                    random_state=rng.randint(0, 2**31 - 1),
                )
            else:
                self.best_model_ = TabPFNRegressor(
                    device=self.device,
                    random_state=rng.randint(0, 2**31 - 1),
                )
            self.best_model_.fit(X, y)


class TunedTabPFNClassifier(TunedTabPFNBase, ClassifierMixin):
    """TabPFN Classifier with hyperparameter tuning and proper categorical handling."""

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        categorical_feature_indices: Optional[list[int]] = None,
    ) -> TunedTabPFNClassifier:
        # Validate input
        X, y = check_X_y(
            X,
            y,
            force_all_finite="allow-nan",
            dtype=object,
            accept_sparse=False,
        )

        # Store dimensions
        self.n_features_in_ = X.shape[1]

        # Set up encoders
        self._setup_data_encoders(X, categorical_feature_indices)
        self._label_encoder = LabelEncoder()
        y = self._label_encoder.fit_transform(y)

        # Store classes
        self.classes_ = self._label_encoder.classes_

        # Determine task type and optimize
        task_type = "multiclass" if len(self.classes_) > 2 else "binary"
        self._optimize(X, y, task_type)

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        check_is_fitted(self)
        X = check_array(X, force_all_finite="allow-nan", dtype=object)
        X = check_array(
            self._cat_encoder.transform(X),
            force_all_finite="allow-nan",
            dtype="numeric",
        )
        return self._label_encoder.inverse_transform(self.best_model_.predict(X))

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        check_is_fitted(self)
        X = check_array(X, force_all_finite="allow-nan", dtype=object)
        X = check_array(
            self._cat_encoder.transform(X),
            force_all_finite="allow-nan",
            dtype="numeric",
        )
        return self.best_model_.predict_proba(X)

    def _more_tags(self):
        return {"allow_nan": True}


class TunedTabPFNRegressor(TunedTabPFNBase, RegressorMixin):
    """TabPFN Regressor with hyperparameter tuning and proper categorical handling."""

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        categorical_feature_indices: Optional[list[int]] = None,
    ) -> TunedTabPFNRegressor:
        # Validate input
        X, y = check_X_y(
            X,
            y,
            force_all_finite="allow-nan",
            dtype=object,
            accept_sparse=False,
        )

        # Store dimensions
        self.n_features_in_ = X.shape[1]

        # Set up encoders
        self._setup_data_encoders(X, categorical_feature_indices)

        # Optimize
        self._optimize(X, y, "regression")

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        check_is_fitted(self)
        X = check_array(X, force_all_finite="allow-nan", dtype=object)
        X = check_array(
            self._cat_encoder.transform(X),
            force_all_finite="allow-nan",
            dtype="numeric",
        )
        return self.best_model_.predict(X)

    def _more_tags(self):
        return {"allow_nan": True}



import pytest
import os
import importlib.util
from pathlib import Path
from typing import Generator, List


def get_example_files() -> List[Path]:
    """Get all Python files from the examples directory."""
    package_root = Path(__file__).parent.parent
    examples_dir = package_root / "examples"
    return list(examples_dir.glob("**/*.py"))


def import_module_from_path(path: Path) -> object:
    """Dynamically import a Python module from a file path."""
    spec = importlib.util.spec_from_file_location(path.stem, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


@pytest.fixture
def example_files() -> Generator[List[Path], None, None]:
    """Fixture that provides all example files."""
    files = get_example_files()
    if not files:
        pytest.skip("No example files found")
    yield files


@pytest.mark.parametrize("example_file", get_example_files(), ids=lambda x: x.name)
def test_example(example_file):
    """Test that each example file can be imported."""
    try:
        import_module_from_path(example_file)
    except Exception as e:
        pytest.fail(f"Failed to import {example_file}: {str(e)}")


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

import numpy as np
from sklearn.feature_selection import SequentialFeatureSelector
from sklearn.model_selection import cross_val_score


def feature_selection(
    estimator, X, y, n_features_to_select=3, feature_names=None, **kwargs
):
    if hasattr(estimator, "fit_at_predict_time"):
        if estimator.fit_at_predict_time:
            print(
                "WARNING: We recommend to set fit_at_predict_time to False for SHAP values to "
                "be calculated, this will significantly speed up calculation."
            )

    if hasattr(estimator, "show_progress"):
        show_progress_ = estimator.show_progress
        estimator.show_progress = False
        try:
            return _feature_selection(
                estimator, X, y, n_features_to_select, feature_names, **kwargs
            )
        finally:
            estimator.show_progress = show_progress_
    else:
        return _feature_selection(
            estimator, X, y, n_features_to_select, feature_names, **kwargs
        )


def _feature_selection(
    estimator, X, y, n_features_to_select=3, feature_names=None, **kwargs
):
    # TODO: Try https://rasbt.github.io/mlxtend/api_subpackages/mlxtend.feature_selection/#sequentialfeatureselector
    # TODO: Could use more feature in training, but only keep fewer in test
    # TODO: the fit function is somehow still called; We need to to change the feature selection
    #  method so it sets feature to none in test; Could be done with a wrapper that sets the feature subsets in fit
    #  and predict
    CV_FOLDS = 5

    score_all = cross_val_score(estimator, X, y, cv=CV_FOLDS)

    # TODO: Feature selection is done without CV, i.e. final CV scores might be biased (too good)
    sfs = SequentialFeatureSelector(
        estimator, n_features_to_select=n_features_to_select, direction="forward"
    )
    sfs.fit(X, y)
    sfs.get_support()
    X_transformed = sfs.transform(X)

    score_selected = cross_val_score(estimator, X_transformed, y, cv=CV_FOLDS)

    print(f"Score with all features: {score_all.mean()} +/- {score_all.std()}")
    print(
        f"Score with selected features: {score_selected.mean()} +/- {score_selected.std()}"
    )

    if feature_names is not None:
        print(
            "Features selected by forward sequential selection: "
            f"{np.array(feature_names)[sfs.get_support()].tolist()}"
        )

    return sfs



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

import numpy as np
from typing import Dict, List, Literal, Optional, Tuple


def infer_categorical_features(X: np.ndarray, categorical_features) -> List[int]:
    """
    Infer the categorical features from the input data.
    We take `self.categorical_features` as the initial list of categorical features.


    Parameters:
        X (ndarray): The input data.

    Returns:
        Tuple[int, ...]: The indices of the categorical features.
    """
    max_unique_values_as_categorical_feature = 10
    min_unique_values_as_numerical_feature = 10

    _categorical_features = []
    for i in range(X.shape[-1]):
        # Filter categorical features, with too many unique values
        if i in categorical_features and (
            len(np.unique(X[:, i])) <= max_unique_values_as_categorical_feature
        ):
            _categorical_features += [i]

        # Filter non-categorical features, with few unique values
        elif (
            i not in categorical_features
            and len(np.unique(X[:, i])) < min_unique_values_as_numerical_feature
            and X.shape[0] > 100
        ):
            categorical_features += [i]

    return _categorical_features



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

import numpy as np
import torch
from sklearn.base import RegressorMixin
from sklearn.preprocessing import LabelEncoder

from tabpfn_extensions import TabPFNRegressor
from tabpfn.model.bar_distribution import BarDistribution


class ClassifierAsRegressor(RegressorMixin):
    """Wrapper class to use a classifier as a regressor.

    This class takes a classifier estimator and converts it into a regressor by
    encoding the target labels and treating the regression problem as a
    classification task.

    Parameters:
        estimator : object
            Classifier estimator to be used as a regressor.

    Attributes:
        label_encoder_ : LabelEncoder
            Label encoder used to transform target regression labels to classes.
        y_train_ : array-like of shape (n_samples,)
            Transformed target labels used for training.
        categorical_features : list
            List of categorical feature indices.

    Examples:
    ```python title="Example"
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import train_test_split
    >>> from tabpfn_extensions import ManyClassClassifier, TabPFNClassifier, ClassifierAsRegressor
    >>> x, y = load_diabetes(return_X_y=True)
    >>> x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=42)
    >>> clf = TabPFNClassifier()
    >>> clf = ManyClassClassifier(clf, n_estimators=10, alphabet_size=clf.max_num_classes_)
    >>> reg = ClassifierAsRegressor(clf)
    >>> reg.fit(x_train, y_train)
    >>> y_pred = reg.predict(x_test)
    ```
    """

    def __init__(self, estimator):
        assert hasattr(
            estimator, "predict_proba"
        ), "The estimator must have a predict_proba method to be used as a regressor."
        self.estimator = estimator

    def fit(self, X, y):
        """Fit the classifier as a regressor.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Training data.
            y : array-like of shape (n_samples,)
                Target labels.

        Returns:
            self : object
                Fitted estimator.
        """
        self.label_encoder_ = LabelEncoder()
        y_transformed = self.label_encoder_.fit_transform(y)
        self.y_train_ = y_transformed

        return self.estimator.fit(X, y_transformed)

    def _get_bar_dist(self):
        """Get the bar distribution for the target labels.

        Returns:
            BarDistribution
                Bar distribution for the target labels.
        """
        unique_y = self.label_encoder_.classes_
        bucket_widths = unique_y[1:] - unique_y[:-1]

        borders = (
            unique_y - (np.array([bucket_widths[0]] + bucket_widths.tolist()) / 2)
        ).tolist() + [unique_y[-1] + bucket_widths[-1] / 2]
        return BarDistribution(borders=torch.tensor(borders).float())

    def predict(self, X):
        """Predict the target values for the input data.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Input data.

        Returns:
            y_pred : array-like of shape (n_samples,)
                Predicted target values.
        """
        return self.predict_full(X)["mean"]

    def set_categorical_features(self, categorical_features):
        """Set the categorical feature indices.

        Parameters:
            categorical_features : list
                List of categorical feature indices.
        """
        self.categorical_features = categorical_features

    def get_optimization_mode(self):
        """Get the optimization mode for the regressor.

        Returns:
            str
                Optimization mode ("mean").
        """
        return "mean"

    @staticmethod
    def probabilities_to_logits_multiclass(probabilities, eps=1e-6):
        """Convert probabilities to logits for a multi-class problem.

        Parameters:
            probabilities : array-like of shape (n_samples, n_classes)
                Input probabilities for each class.
            eps : float, default=1e-6
                Small value to avoid division by zero or taking logarithm of zero.

        Returns:
            logits : array-like of shape (n_samples, n_classes)
                Output logits for each class.
        """
        probabilities = np.clip(probabilities, eps, 1 - eps)
        logits = np.log(probabilities) - np.log(
            1 - probabilities.sum(axis=1, keepdims=True) + eps
        )
        return logits

    def predict_full(self, X):
        """Predict the full set of output values for the input data.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Input data.

        Returns:
            dict
                Dictionary containing the predicted output values, including:
                - "mean": Predicted mean values.
                - "median": Predicted median values.
                - "mode": Predicted mode values.
                - "logits": Predicted logits.
                - "buckets": Predicted bucket probabilities.
                - "quantile_{q:.2f}": Predicted quantile values for each quantile q.
        """
        class_probs = self.estimator.predict_proba(X)
        class_logits = (
            torch.tensor(
                ClassifierAsRegressor.probabilities_to_logits_multiclass(class_probs)
            )
            .float()
            .unsqueeze(0)
        )
        bar_dist = self._get_bar_dist()
        r = TabPFNRegressor._post_process_predict_full(
            prediction=class_logits, criterion=bar_dist
        )

        return r



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

import math
import numpy as np
from sklearn.multiclass import OutputCodeClassifier
from sklearn.utils import check_random_state
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import _num_samples, check_is_fitted


class ManyClassClassifier(OutputCodeClassifier):
    """Output-Code multiclass strategy with deciary codebook.

    This class extends the original OutputCodeClassifier to support n-ary codebooks (with n=alphabet_size),
    allowing for handling more classes.

    Parameters:
        estimator : estimator object
            An estimator object implementing :term:`fit` and one of
            :term:`decision_function` or :term:`predict_proba`. The base classifier
            should be able to handle up to `alphabet_size` classes.

        random_state : int, RandomState instance, default=None
            The generator used to initialize the codebook.
            Pass an int for reproducible output across multiple function calls.
            See :term:`Glossary <random_state>`.

    Attributes:
        estimators_ : list of `int(n_classes * code_size)` estimators
            Estimators used for predictions.

        classes_ : ndarray of shape (n_classes,)
            Array containing labels.

        code_book_ : ndarray of shape (n_classes, `len(estimators_)`)
            Deciary array containing the code of each class.

    Examples:
    ```python title="Example"
    >>> from sklearn.datasets import load_iris
    >>> from tabpfn.scripts.estimator import ManyClassClassifier, TabPFNClassifier
    >>> from sklearn.model_selection import train_test_split
    >>> x, y = load_iris(return_X_y=True)
    >>> x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=42)
    >>> clf = TabPFNClassifier()
    >>> clf = ManyClassClassifier(clf, alphabet_size=clf.max_num_classes_)
    >>> clf.fit(x_train, y_train)
    >>> clf.predict(x_test)
    ```
    """

    def __init__(
        self,
        estimator,
        *,
        alphabet_size=None,
        n_estimators=None,
        n_estimators_redundancy=4,
        random_state=None,
    ):
        self.estimator = estimator
        self.random_state = random_state
        self.alphabet_size = alphabet_size
        self.n_estimators = n_estimators
        self.n_estimators_redundancy = n_estimators_redundancy

    def get_alphabet_size(self):
        if self.alphabet_size is None:
            return self.estimator.max_num_classes_
        return self.alphabet_size

    def get_n_estimators(self, n_classes):
        if self.n_estimators is None:
            return (
                math.ceil(1 + math.log(n_classes, self.get_alphabet_size()))
                * 2
                * self.n_estimators_redundancy
            )
        return self.n_estimators

    def _generate_codebook(self, n_classes, n_estimators, alphabet_size):
        """Generate an efficient codebook using the provided alphabet size.

        This function generates a codebook where for each codeword at most `alphabet_size - 1` classes
        are mapped to unique classes, and the remaining classes are mapped to the "rest" class. The codebook
        is generated to provide for each class a unique codeword with maximum reduncancy between codewords.
        Greedy optimization is used to find the optimal codewords.

        Parameters:
            n_classes : int
                Number of classes.
            n_estimators : int
                Number of estimators.
            alphabet_size : int, default=10
                Size of the alphabet for the codebook. The first `alphabet_size - 1` classes
                will be mapped to unique classes, and the remaining classes will be mapped
                to the "rest" class.

        Returns:
            codebook : ndarray of shape (n_estimators, n_classes)
                Efficient n-ary codebook.
        """
        n_classes_in_codebook = min(n_classes, alphabet_size - 1)
        n_classes_remaining = n_classes - n_classes_in_codebook

        # Initialize the codebook with zeros
        codebook = np.zeros((n_estimators, n_classes), dtype=int)

        def generate_codeword():
            choices = list(range(0, n_classes_in_codebook)) + [
                n_classes_in_codebook for _ in range(n_classes_remaining)
            ]
            return np.random.permutation(
                choices,
            )

        # Generate the first codeword randomly
        codebook[0] = generate_codeword()

        # Generate the remaining codewords
        for i in range(1, n_estimators):
            # Initialize the current codeword with random values
            current_codeword = generate_codeword()
            max_distance, n_min_distance_min, max_distance_codeword = (
                0,
                1000,
                current_codeword,
            )

            # Iteratively improve the current codeword
            for _ in range(1000):  # Number of iterations can be adjusted
                # Compute the Hamming distances between the current codeword and all previous codewords
                distances = np.sum(
                    (codebook[:i] == n_classes_in_codebook)
                    != (current_codeword == n_classes_in_codebook),
                    axis=0,
                )

                # Find the minimum Hamming distance
                min_distance = np.min(distances, axis=0)

                n_min_distance = np.sum(distances == min_distance)

                # print(min_distance, n_min_distance)

                if min_distance > max_distance or (
                    min_distance == max_distance and n_min_distance < n_min_distance_min
                ):
                    max_distance_codeword = current_codeword
                    max_distance = min_distance
                    n_min_distance_min = n_min_distance

                current_codeword = generate_codeword()

            # Assign the optimized codeword to the codebook
            codebook[i] = max_distance_codeword

        return codebook

    def fit(self, X, y, **fit_params):
        """Fit underlying estimators.

        Parameters:
            X : {array-like, sparse matrix} of shape (n_samples, n_features)
                Data.

            y : array-like of shape (n_samples,)
                Multi-class targets.

            **fit_params : dict
                Parameters passed to the ``estimator.fit`` method of each
                sub-estimator.

        Returns:
            self : object
                Returns a fitted instance of self.
        """
        y = self._validate_data(X="no_validation", y=y)

        random_state = check_random_state(self.random_state)
        check_classification_targets(y)

        self.classes_ = np.unique(y)
        n_classes = self.classes_.shape[0]
        if n_classes == 0:
            raise ValueError(
                "DeciaryOutputCodeClassifier can not be fit when no class is present."
            )

        self.no_mapping_needed_ = False
        if n_classes <= self.get_alphabet_size():
            self.no_mapping_needed_ = True
            return self.estimator.fit(X, y, **fit_params)

        # Generate efficient deciary codebook
        self.code_book_ = self._generate_codebook(
            n_classes,
            self.get_n_estimators(n_classes),
            alphabet_size=self.get_alphabet_size(),
        )

        print(
            f"Using codebook with {self.code_book_.shape[0]} estimators and {self.code_book_.shape[1]} classes"
        )

        self.classes_index_ = {c: i for i, c in enumerate(self.classes_)}

        self.Y_train = np.array(
            [self.code_book_[:, y[i]] for i in range(_num_samples(y))],
            dtype=int,
        )
        self.X_train = X

        # print('CLF DIST IN', list(zip(np.unique(y), (np.unique(y, return_counts=True)[1] / y.shape[0]).tolist())))

        return self

    def predict_proba(self, X):
        """Predict probabilities using the underlying estimators.

        Parameters:
            X : {array-like, sparse matrix} of shape (n_samples, n_features)
                Data.

        Returns:
            p : ndarray of shape (n_samples, n_classes)
                Returns the probability of the samples for each class in the model,
                where classes are ordered as they are in `self.classes_`.
        """
        check_is_fitted(self)

        if self.no_mapping_needed_:
            return self.estimator.predict_proba(X)

        Y = np.array(
            [
                _fit_and_predict_proba(
                    self.estimator, self.X_train, self.Y_train[:, i], X
                )
                for i in range(self.code_book_.shape[0])
            ],
            dtype=np.float64,
        )

        # Y shape is (n_estimators, n_samples, n_classes_per_code)
        n_estimators, n_samples, n_classes_per_code = Y.shape
        rest_class = np.max(self.code_book_)

        # Compute the weighted probabilities for each class
        probabilities = np.zeros((n_samples, self.classes_.shape[0]))

        for i in range(n_estimators):
            for j in range(self.classes_.shape[0]):
                if self.code_book_[i, j] != rest_class:
                    j_remapped = self.code_book_[i, j]
                    probabilities[:, j] += Y[
                        i, :, j_remapped
                    ]  # / (1 - Y[i, :, rest_class])
                    # print(j, Y[i, :, j_remapped])

        assert not (
            (self.code_book_ != rest_class).sum(0) == 0
        ).any(), f"Some classes are not mapped to any codeword. {self.code_book_} {self.classes_} {((self.code_book_ != rest_class).sum(0) == 0)}"

        # Normalize the weighted probabilities to get the final class probabilities
        probabilities /= (self.code_book_ != rest_class).sum(0)
        probabilities /= probabilities.sum(axis=1, keepdims=True)

        return probabilities

    def set_categorical_features(self, categorical_features):
        self.categorical_features = categorical_features


def _fit_and_predict_proba(estimator, X_train, Y_train, X):
    """Predict probabilities for deciary values using a single estimator."""
    estimator.fit(X_train, Y_train)

    if hasattr(estimator, "predict_proba"):
        return estimator.predict_proba(X)
    else:
        raise AttributeError("Base estimator doesn't have `predict_proba` method.")



import numpy as np
import pytest
from tabpfn_extensions.rf_pfn import (
    RandomForestTabPFNClassifier,
    DecisionTreeTabPFNClassifier,
    RandomForestTabPFNRegressor,
    DecisionTreeTabPFNRegressor,
)
from tabpfn_extensions import TabPFNClassifier, TabPFNRegressor
from sklearn.utils.estimator_checks import check_estimator, parametrize_with_checks

@pytest.mark.parametrize("model_class", [
    (RandomForestTabPFNClassifier, TabPFNClassifier),
    (RandomForestTabPFNRegressor, TabPFNRegressor),
    (DecisionTreeTabPFNClassifier, TabPFNClassifier),
    (DecisionTreeTabPFNRegressor, TabPFNRegressor)
])
def test_sklearn_compatibility(model_class):
    """Test RandomForestTabPFN compatibility with different sklearn versions."""
    # Generate sample data
    rng = np.random.RandomState(42)
    X = rng.rand(100, 4)
    y = rng.randint(0, 2, 100)
    
    # Initialize classifier
    clf_class, clf_base_class = model_class
    clf_base = clf_base_class()
    kwargs = {
        "tabpfn": clf_base,
        **({"n_estimators": 2} if "RandomForest" in clf_class.__name__ else {}),  # Only add n_estimators for RandomForest
        "max_depth": 2
    }
    clf = clf_class(**kwargs)
    
    # This should work without errors on both sklearn <1.6 and >=1.6
    clf.fit(X, y)
    
    # Verify predictions work
    pred = clf.predict(X)
    assert pred.shape == (100,)
    if clf_base_class == TabPFNClassifier:
        assert np.all(np.isin(pred, [0, 1]))
    else:
        assert pred.dtype == np.float64

@pytest.mark.parametrize("model_class", [
    (RandomForestTabPFNClassifier, TabPFNClassifier),
    (RandomForestTabPFNRegressor, TabPFNRegressor),
    (DecisionTreeTabPFNClassifier, TabPFNClassifier),
    (DecisionTreeTabPFNRegressor, TabPFNRegressor)
])
def test_with_nan(model_class):
    """Test handling of NaN values for sklearn < 1.6."""
    # Generate sample data with NaN values
    rng = np.random.RandomState(42)
    X = rng.rand(100, 4)
    X[0, 0] = np.nan  # Add a NaN value
    y = rng.randint(0, 2, 100)
    
    clf_class, clf_base_class = model_class
    clf_base = clf_base_class()
    kwargs = {
        "tabpfn": clf_base,
        **({"n_estimators": 2} if "RandomForest" in clf_class.__name__ else {}),
        "max_depth": 2
    }
    clf = clf_class(**kwargs)
    
    # This should work without errors
    clf.fit(X, y)
    
    # Test prediction with NaN values
    X_test = X.copy()
    pred = clf.predict(X_test)
    
    assert pred.shape == (100,)
    if clf_base_class == TabPFNClassifier:
        assert np.all(np.isin(pred, [0, 1]))
    else:
        assert pred.dtype == np.float64



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

import os

import os
from typing import Any, Type, Tuple, Protocol, Literal, Optional, Union, Dict
from dataclasses import dataclass
from typing_extensions import override
from sklearn.base import BaseEstimator
import numpy as np
import warnings

class TabPFNEstimator(Protocol):
    def fit(self, X: Any, y: Any) -> Any:
        ...

    def predict(self, X: Any) -> Any:
        ...


def is_tabpfn(estimator: Any) -> bool:
    """Check if an estimator is a TabPFN model."""
    try:
        return any(
            [
                "TabPFN" in str(estimator.__class__),
                "TabPFN" in str(estimator.__class__.__bases__),
                any("TabPFN" in str(b) for b in estimator.__class__.__bases__),
                "tabpfn.base_model.TabPFNBaseModel" in str(estimator.__class__.mro()),
            ]
        )
    except (AttributeError, TypeError):
        return False

from typing import Tuple, Type
import os
USE_TABPFN_LOCAL = os.getenv("USE_TABPFN_LOCAL", "true").lower() == "true"

def get_tabpfn_models() -> Tuple[Type, Type, Type]:
    """Get TabPFN models with fallback between local and client versions."""
    if USE_TABPFN_LOCAL:
        try:
            from tabpfn import TabPFNClassifier, TabPFNRegressor

            return TabPFNClassifier, TabPFNRegressor
        except ImportError:
            pass

    try:
        from tabpfn_client import (
            TabPFNClassifier as ClientTabPFNClassifier,
            TabPFNRegressor as ClientTabPFNRegressor,
        )


        # Wrapper classes to add device parameter
        # we can't use *args because scikit-learn needs to know the parameters of the constructor
        class TabPFNClassifierWrapper(ClientTabPFNClassifier):
            def __init__(
                self,
                device: Union[str, None] = None,
                categorical_features_indices: Optional[list[int]] = None,
                model_path: str = "default",
                n_estimators: int = 4,
                softmax_temperature: float = 0.9,
                balance_probabilities: bool = False,
                average_before_softmax: bool = False,
                ignore_pretraining_limits: bool = False,
                inference_precision: Literal["autocast", "auto"] = "auto",
                random_state: Optional[Union[int, np.random.RandomState, np.random.Generator]] = None,
                inference_config: Optional[Dict] = None,
                paper_version: bool = False,
            ) -> None:
                self.device = device
                #TODO: we should support this argument in the client version
                self.categorical_features_indices = categorical_features_indices
                if categorical_features_indices is not None:
                    warnings.warn(
                        "categorical_features_indices is not supported in the client version of TabPFN and will be ignored",
                        UserWarning,
                        stacklevel=2,
                    )
                if "/" in model_path:
                    model_name = model_path.split("/")[-1].split("-")[-1].split(".")[0]
                    if model_name == "classifier":
                        model_name = "default"
                    self.model_path = model_name
                else:
                    self.model_path = model_path

                super().__init__(
                    model_path=self.model_path,
                    n_estimators=n_estimators,
                    softmax_temperature=softmax_temperature,
                    balance_probabilities=balance_probabilities,
                    average_before_softmax=average_before_softmax,
                    ignore_pretraining_limits=ignore_pretraining_limits,
                    inference_precision=inference_precision,
                    random_state=random_state,
                    inference_config=inference_config,
                    paper_version=paper_version,
                )

            def get_params(self, deep: bool = True) -> Dict[str, Any]:
                """Return parameters for this estimator."""
                params = super().get_params(deep=deep)
                params.pop("device")
                params.pop("categorical_features_indices")
                return params

        class TabPFNRegressorWrapper(ClientTabPFNRegressor):
            def __init__(
                self,
                device: Union[str, None] = None,
                categorical_features_indices: Optional[list[int]] = None,
                model_path: str = "default",
                n_estimators: int = 8,
                softmax_temperature: float = 0.9,
                average_before_softmax: bool = False,
                ignore_pretraining_limits: bool = False,
                inference_precision: Literal["autocast", "auto"] = "auto",
                random_state: Optional[Union[int, np.random.RandomState, np.random.Generator]] = None,
                inference_config: Optional[Dict] = None,
                paper_version: bool = False,
            ) -> None:
                self.device = device
                self.categorical_features_indices = categorical_features_indices
                if categorical_features_indices is not None:
                    warnings.warn(
                        "categorical_features_indices is not supported in the client version of TabPFN and will be ignored",
                        UserWarning,
                        stacklevel=2,
                    )

                if "/" in model_path:
                    model_name = model_path.split("/")[-1].split("-")[-1].split(".")[0]
                    if model_name == "regressor":
                        model_name = "default"
                    self.model_path = model_name
                else:
                    self.model_path = model_path

                super().__init__(
                    model_path=self.model_path,
                    n_estimators=n_estimators,
                    softmax_temperature=softmax_temperature,
                    average_before_softmax=average_before_softmax,
                    ignore_pretraining_limits=ignore_pretraining_limits,
                    inference_precision=inference_precision,
                    random_state=random_state,
                    inference_config=inference_config,
                    paper_version=paper_version,
                )

            def get_params(self, deep: bool = True) -> Dict[str, Any]:
                """
                Return parameters for this estimator.
                """
                params = super().get_params(deep=deep)
                params.pop("device")
                params.pop("categorical_features_indices")
                return params

        return TabPFNClassifierWrapper, TabPFNRegressorWrapper

    except ImportError:
        raise ImportError(
            "Neither local TabPFN nor TabPFN client could be imported. Install with:\n"
            "pip install tabpfn\n"
            "or\n"
            "pip install tabpfn-client"
        )


TabPFNClassifier, TabPFNRegressor = get_tabpfn_models()



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from multiprocessing import Pool

import shap
from ..utils import is_tabpfn


def calculate_shap_subset(args):
    X_subset, background, model, feature_idx = args
    explainer = shap.PermutationExplainer(model, background)
    return explainer.shap_values(X_subset)[:, feature_idx]


def parallel_permutation_shap(model, X, background=None, n_jobs=-1):
    # If no background data provided, use X
    if background is None:
        background = X

    # Split data into chunks for parallel processing
    n_features = X.shape[1]
    feature_indices = list(range(n_features))

    # Prepare arguments for parallel processing
    args_list = [(X, background, model, idx) for idx in feature_indices]

    # Run parallel computation
    with Pool(processes=n_jobs) as pool:
        shap_values_per_feature = pool.map(calculate_shap_subset, args_list)

    # Combine results
    shap_matrix = np.column_stack(shap_values_per_feature)

    return shap_matrix


def plot_shap(shap_values: np.ndarray):
    """
    Plots the shap values for the given test data. It will plot aggregated shap values for each feature, as well
    as per sample shap values. Additionally, if multiple samples are provided, it will plot the 3 most important
    interactions with the most important feature.

    Parameters:
        shap_values:
    """
    import shap

    if len(shap_values.shape) == 3:
        print("Computing shap values for the first class (index 0).")
        shap_values = shap_values[:, :, 0]

    shap.plots.bar(shap_values=shap_values, show=False)
    plt.title("Aggregate feature importances across the test examples")
    plt.show()
    shap.summary_plot(shap_values=shap_values, show=False)
    # plot the distribution of importances for each feature over all samples
    plt.title(
        "Feature importances for each feature for each test example (a dot is one feature for one example)"
    )
    plt.show()

    most_important = shap_values.abs.mean(0).values.argsort()[-1]
    print(
        f'Now we analyze the strongest feature interactions of the most important feature, namely the feature "{most_important}".'
    )
    if len(shap_values) > 1:
        plot_shap_feature(shap_values, most_important)


def plot_shap_feature(shap_values_, feature_name, n_plots=1):
    import shap

    # we can use shap.approximate_interactions to guess which features
    # may interact with age
    inds = shap.utils.potential_interactions(
        shap_values_[:, feature_name], shap_values_
    )

    # make plots colored by each of the top three possible interacting features
    for i in range(n_plots):
        shap.plots.scatter(
            shap_values_[:, feature_name],
            color=shap_values_[:, inds[i]],
            show=False,
        )
        plt.title(
            f"Feature {feature_name} with a color coding representing the value of ({inds[i]})"
        )


def get_shap_values(estimator, test_x, attribute_names=None, **kwargs) -> np.ndarray:
    """
    Computes SHAP (SHapley Additive exPlanations) values for the model's predictions on the given input features.

    Parameters:
        test_x (Union[pd.DataFrame, np.ndarray]): The input features to compute SHAP values for.
        kwargs (dict): Additional keyword arguments to pass to the SHAP explainer.

    Returns:
        np.ndarray: The computed SHAP values.
    """

    if isinstance(test_x, torch.Tensor):
        test_x = test_x.cpu().numpy()

    if isinstance(test_x, np.ndarray):
        test_x = pd.DataFrame(test_x)
        if attribute_names is not None:
            test_x.rename(
                columns={c: n for c, n in enumerate(attribute_names.tolist())},
                inplace=True,
            )
        else:
            test_x.rename(columns={c: str(c) for c in test_x.columns}, inplace=True)

    if hasattr(estimator, "predict_function_for_shap"):
        predict_function_for_shap = estimator.predict_function_for_shap
    else:
        predict_function_for_shap = (
            "predict_proba" if hasattr(estimator, "predict_proba") else "predict"
        )

    if hasattr(estimator, "fit_at_predict_time"):
        if not estimator.fit_at_predict_time:
            print(
                "WARNING: We recommend to set fit_at_predict_time to True for SHAP values to "
                "be calculated, this will significantly speed up calculation."
            )

    def get_shap():
        if is_tabpfn(estimator):
            explainer = get_tabpfn_explainer(
                estimator, test_x, predict_function_for_shap, **kwargs
            )
        else:
            explainer = get_default_explainer(
                estimator, test_x, predict_function_for_shap, **kwargs
            )
        return explainer(test_x)

    if hasattr(estimator, "show_progress"):
        show_progress_ = estimator.show_progress
        estimator.show_progress = False
        try:
            shap_values = get_shap()
        finally:
            estimator.show_progress = show_progress_
    else:
        shap_values = get_shap()

    return shap_values


def get_tabpfn_explainer(
    estimator, test_x, predict_function_for_shap="predict", **kwargs
):
    import shap

    return shap.Explainer(
        getattr(estimator, predict_function_for_shap),
        np.ones(test_x.iloc[0:1, :].shape) * float("nan"),
        **kwargs,
    )


def get_default_explainer(
    estimator, test_x, predict_function_for_shap="predict", **kwargs
):
    import shap

    background = shap.maskers.Independent(test_x, max_samples=1000)

    return shap.Explainer(
        getattr(estimator, predict_function_for_shap),
        test_x,
        **kwargs,
    )



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import copy
import numpy as np
import random
import torch
from sklearn.base import BaseEstimator
from tqdm import tqdm
from typing import Optional

from .. import utils_todo
from tabpfn_extensions import TabPFNRegressor, TabPFNClassifier
from tabpfn_extensions.utils import USE_TABPFN_LOCAL


class TabPFNUnsupervisedModel(BaseEstimator):
    """TabPFN experiments model for imputation, outlier detection, and synthetic data generation.

    This model combines a TabPFNClassifier for categorical features and a TabPFNRegressor for
    numerical features to perform various experiments learning tasks on tabular data.

    Parameters:
        tabpfn_clf : TabPFNClassifier, optional
            TabPFNClassifier instance for handling categorical features. If not provided, the model
            assumes that there are no categorical features in the data.

        tabpfn_reg : TabPFNRegressor, optional
            TabPFNRegressor instance for handling numerical features. If not provided, the model
            assumes that there are no numerical features in the data.

    Attributes:
        categorical_features : list
            List of indices of categorical features in the input data.

    Examples:
    ```python title="Example"
    >>> tabpfn_clf = TabPFNClassifier()
    >>> tabpfn_reg = TabPFNRegressor()
    >>> model = TabPFNUnsupervisedModel(tabpfn_clf, tabpfn_reg)
    >>>
    >>> X = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]
    >>> model.fit(X)
    >>>
    >>> X_imputed = model.impute(X)
    >>> X_outliers = model.outliers(X)
    >>> X_synthetic = model.generate_synthetic_data(n_samples=100)
    ```
    """

    def _more_tags(self):
        return {"allow_nan": True}

    def __init__(
        self,
        tabpfn_clf: Optional[TabPFNClassifier] = None,
        tabpfn_reg: Optional[TabPFNRegressor] = None,
    ) -> None:
        """Initialize the TabPFNUnsupervisedModel.

        Args:
            tabpfn_clf : TabPFNClassifier, optional
                TabPFNClassifier instance for handling categorical features. If not provided, the model
                assumes that there are no categorical features in the data.

            tabpfn_reg : TabPFNRegressor, optional
                TabPFNRegressor instance for handling numerical features. If not provided, the model
                assumes that there are no numerical features in the data.

        Raises:
            AssertionError
                If both tabpfn_clf and tabpfn_reg are None.
        """
        assert (
            tabpfn_clf is not None or tabpfn_reg is not None
        ), "You cannot set both `tabpfn_clf` and `tabpfn_reg` to None. You can set one to None, if your table exclusively consists of categoricals/numericals."

        self.tabpfn_clf = tabpfn_clf
        self.tabpfn_reg = tabpfn_reg
        self.estimators = [self.tabpfn_clf, self.tabpfn_reg]

        self.categorical_features = []

    def set_categorical_features(self, categorical_features):
        self.categorical_features = categorical_features
        for estimator in self.estimators:
            try:
                estimator.set_categorical_features(categorical_features)
            except:
                pass

    def fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> None:
        """Fit the model to the input data.

        Args:
            X : array-like of shape (n_samples, n_features)
                Input data to fit the model.

            y : array-like of shape (n_samples,), optional
                Target values.

        Returns:
            self : TabPFNUnsupervisedModel
                Fitted model.
        """
        self.X_ = copy.deepcopy(X)
        self.y = copy.deepcopy(y)
        self.categorical_features = utils_todo.infer_categorical_features(
            X, self.categorical_features
        )

    def init_model_and_get_model_config(self):
        for estimator in self.estimators:
            estimator.init_model_and_get_model_config()

    def impute_(
        self,
        X: torch.tensor,
        t: float = 0.000000001,
        n_permutations: int = 10,
        condition_on_all_features: bool = True,
    ) -> torch.tensor:
        """
        Impute missing values (np.nan) in X by sampling all cells independently from the trained models

        :param X: Input data of the shape (num_examples, num_features) with missing values encoded as np.nan
        :param t: Temperature for sampling from the imputation distribution, lower values are more deterministic
        :return: Imputed data, with missing values replaced
        """

        n_features = X.shape[1]
        all_features = list(range(n_features))

        X_fit = self.X_
        impute_X = copy.deepcopy(X)

        for i in tqdm(range(len(all_features))):
            column_idx = all_features[i]

            if not condition_on_all_features:
                conditional_idx = all_features[:i] if i > 0 else []
            else:
                conditional_idx = list(set(range(X.shape[1])) - {column_idx})

            y_predict = impute_X[:, column_idx]

            if torch.isnan(y_predict).sum() == 0:
                continue

            X_where_y_is_nan = impute_X[torch.isnan(y_predict)]
            X_where_y_is_nan = X_where_y_is_nan.reshape(-1, impute_X.shape[1])

            densities = []
            for perm in efficient_random_permutation(conditional_idx, n_permutations):
                perm = perm + (column_idx,)
                _, pred = self.impute_single_permutation_(
                    X_where_y_is_nan, perm, t, condition_on_all_features
                )
                densities.append(pred)

            if not self.use_classifier_(column_idx, X_fit[:, column_idx]):
                pred_merged = densities[0][
                    "criterion"
                ].average_bar_distributions_into_this(
                    [d["criterion"] for d in densities],
                    [torch.tensor(d["logits"]) for d in densities],
                )
                pred_sampled = densities[0]["criterion"].sample(pred_merged, t=t)
            else:
                pred = torch.stack([d for d in densities]).mean(dim=0)
                pred_sampled = (
                    torch.distributions.Categorical(probs=pred).sample().float()
                )

            impute_X[torch.isnan(y_predict), column_idx] = pred_sampled

        return impute_X

    def impute_single_permutation_(
        self,
        X: torch.tensor,
        feature_permutation: list[int] | tuple[int],
        t: float = 0.000000001,
        condition_on_all_features: bool = True,
    ) -> torch.tensor:
        """
        Impute missing values (np.nan) in X by sampling all cells independently from the trained models

        :param X: Input data of the shape (num_examples, num_features) with missing values encoded as np.nan
        :param t: Temperature for sampling from the imputation distribution, lower values are more deterministic
        :return: Imputed data, with missing values replaced
        """
        X_fit = self.X_
        impute_X = copy.deepcopy(X)

        for i in range(len(feature_permutation)):
            column_idx = feature_permutation[i]

            if not condition_on_all_features:
                conditional_idx = feature_permutation[:i] if i > 0 else []
            else:
                conditional_idx = list(set(range(X.shape[1])) - {column_idx})

            y_predict = impute_X[:, column_idx]

            if torch.isnan(y_predict).sum() == 0:
                continue

            X_where_y_is_nan = impute_X[torch.isnan(y_predict)]
            X_where_y_is_nan = X_where_y_is_nan.reshape(-1, impute_X.shape[1])

            model, X_predict, _ = self.density_(
                X_where_y_is_nan, X_fit, conditional_idx, column_idx
            )

            pred, pred_sampled = self.sample_from_model_prediction_(
                column_idx, X_fit, model, X_predict, t
            )

            impute_X[torch.isnan(y_predict), column_idx] = pred_sampled

        return impute_X, pred

    def sample_from_model_prediction_(self, column_idx, X_fit, model, X_predict, t):
        if not self.use_classifier_(column_idx, X_fit[:, column_idx]):
            pred = model.predict(X_predict.numpy(), output_type="full")
            pred_sampled = pred["criterion"].sample(torch.tensor(pred["logits"]), t=t)
        else:
            pred = model.predict_proba(X_predict.numpy())
            pred_sampled = (
                torch.distributions.Categorical(probs=torch.tensor(pred))
                .sample()
                .float()
            )

        return pred, pred_sampled

    def use_classifier_(self, column_idx, y):
        return (
            column_idx in self.categorical_features
            and len(np.unique(y)) < 10  # TODO: self.tabpfn_clf.max_num_classes_
        )

    def density_(
        self,
        X_predict: torch.tensor,
        X_fit: torch.tensor,
        conditional_idx: list[int],
        column_idx: int,
    ) -> torch.tensor:
        # self.init_model_and_get_model_config()

        if len(conditional_idx) > 0:
            # If not the first feature, use all previous features
            mask = torch.zeros_like(X_fit).bool()
            mask[:, conditional_idx] = True
            X_fit, y_fit = X_fit[mask], X_fit[:, column_idx]
            X_fit = X_fit.reshape(mask.shape[0], -1)

            mask = torch.zeros_like(X_predict).bool()
            mask[:, conditional_idx] = True
            X_predict, y_predict = X_predict[mask], X_predict[:, column_idx]
            X_predict = X_predict.reshape(mask.shape[0], -1)
        else:
            # If the first feature, use a zero feature as input
            # Because of preprocessing, we can't use a zero feature, so we use a random feature
            X_fit, y_fit = torch.randn_like(X_fit[:, 0:1]), X_fit[:, 0]
            X_predict, y_predict = torch.randn_like(X_predict[:, 0:1]), X_predict[:, 0]

        model = (
            self.tabpfn_clf
            if self.use_classifier_(column_idx, y_fit)
            else self.tabpfn_reg
        )

        model.fit(X_fit.numpy(), y_fit.numpy())

        return model, X_predict, y_predict

    def impute(
        self, X: torch.tensor, t: float = 0.000000001, n_permutations: int = 10
    ) -> torch.tensor:
        """Impute missing values in the input data.

        Args:
            X : torch.Tensor of shape (n_samples, n_features)
                Input data with missing values encoded as np.nan.

            t : float, default=0.000000001
                Temperature for sampling from the imputation distribution. Lower values result in
                more deterministic imputations.

        Returns:
            torch.Tensor of shape (n_samples, n_features)
                Imputed data with missing values replaced.
        """
        return self.impute_(
            X, t, condition_on_all_features=True, n_permutations=n_permutations
        )

    def outliers_single_permutation_(
        self, X: torch.tensor, feature_permutation: list[int] | tuple[int]
    ) -> torch.tensor:
        log_p = torch.zeros_like(
            X[:, 0]
        )  # Start with a log probability of 0 (log(1) = 0)

        for i, column_idx in enumerate(feature_permutation):
            model, X_predict, y_predict = self.density_(
                X, self.X_, feature_permutation[:i], column_idx
            )
            if self.use_classifier_(column_idx, y_predict):
                pred = model.predict_proba(X_predict.numpy())[y_predict]
            else:
                pred = model.predict(X_predict.numpy(), output_type="full")
                pred = pred["criterion"].pdf(pred["logits"], torch.tensor(y_predict))

            # Convert probabilities to log probabilities
            log_pred = torch.log(pred)

            # Add log probabilities instead of multiplying probabilities
            log_p = log_p + log_pred

        return log_p, torch.exp(log_p)

    def outliers_pdf(self, X: torch.tensor, n_permutations: int = 10) -> torch.tensor:
        X_store = copy.deepcopy(self.X_)
        mask = torch.ones_like(X_store).bool()
        mask[self.categorical_features] = False
        self.X_ = self.X_[mask]
        mask = torch.ones_like(X).bool()
        mask[self.categorical_features] = False
        X = X[mask]

        pdf = self.outliers(X, n_permutations=n_permutations)
        self.X_ = X_store
        return pdf

    def outliers_pmf(self, X: torch.tensor, n_permutations: int = 10) -> torch.tensor:
        X_store = copy.deepcopy(self.X_)
        mask = torch.zeros_like(X_store).bool()
        mask[self.categorical_features] = True
        self.X_ = self.X_[mask]
        mask = torch.zeros_like(X).bool()
        mask[self.categorical_features] = True
        X = X[mask]

        pmf = self.outliers(X, n_permutations=n_permutations)
        self.X_ = X_store
        return pmf

    def outliers(self, X: torch.tensor, n_permutations: int = 10) -> torch.tensor:
        """
        Preferred implementation for outliers, where we calculate the sample probability for each sample in X by
        multiplying the probabilities of each feature according to chain rule of probability. The first feature is
        estimated by using a zero feature as input.

        Args
            X: Samples to calculate the sample probability for, shape (n_samples, n_features)

        Returns:
            Sample unnormalized probability for each sample in X, shape (n_samples,)
        """
        # self.init_model_and_get_model_config()

        n_features = X.shape[1]
        all_features = list(range(n_features))

        densities = []
        for perm in efficient_random_permutation(all_features, n_permutations):
            perm_density_log, perm_density = self.outliers_single_permutation_(
                X, feature_permutation=perm
            )
            densities.append(perm_density)

        # Average the densities across all permutations
        return torch.stack(densities).mean(dim=0)

    def generate_synthetic_data(self, n_samples=100, t=1.0, n_permutations=3):
        """
        Generate synthetic data using the trained models. Uses imputation method to generate synthetic data, passed with
        a matrix of nans. Samples are generated feature by feature in one pass, so samples are not dependent on each
        other per feature.

        Args:
            n_samples : int, default=100
                Number of synthetic samples to generate.

            t : float, default=1.0
                Temperature for sampling from the imputation distribution. Lower values result in
                more deterministic samples.

        Returns:
            torch.Tensor of shape (n_samples, n_features)
                Generated synthetic data.

        Raises:
            AssertionError
                If the model is not fitted.
        """
        # TODO: Test what happens if we generate one feature at a time, with train data only for that featur
        #  and previous ones, like outliers
        assert hasattr(
            self, "X_"
        ), "You need to fit the model before generating synthetic data"

        X = torch.zeros(n_samples, self.X_.shape[1]) * np.nan
        return self.impute_(
            X, t=t, condition_on_all_features=False, n_permutations=n_permutations
        )

    def get_embeddings(self, X: torch.tensor, per_column: bool = False) -> torch.tensor:
        """
        Get the transformer embeddings for the test data X.

        Args:
            X:

        Returns:
            torch.Tensor of shape (n_samples, embedding_dim)
        """
        raise NotImplementedError(
            "This method is not implemented currently. During the main TabPFN refactor this functionality was removed, please see: https://github.com/PriorLabs/TabPFN/issues/111"
        )

        if per_column:
            return self.get_embeddings_per_column(X)
        else:
            return self.get_embeddings_(X)

    def get_embeddings_(self, X: torch.tensor) -> torch.tensor:
        model = self.tabpfn_reg
        model.fit(
            self.X_,
            self.y
            if self.y is not None
            else (torch.zeros_like(self.X_[:, 0])),  # Must contain more than one class
        )  # Fit the data for random labels
        embs = model.get_embeddings(X, additional_y=None)
        return embs.reshape(X.shape[0], -1)


    def get_embeddings_per_column(self, X: torch.tensor) -> torch.tensor:
        """
        Alternative implementation for get_embeddings, where we get the embeddings for each column as a label
         separately and concatenate the results. This alternative way needs more passes but might be more accurate
        """
        embs = []
        for column_idx in range(0, X.shape[1]):
            mask = torch.zeros_like(self.X_).bool()
            mask[:, column_idx] = True
            X_train, y_train = (
                self.X_[~(mask)].reshape(self.X_.shape[0], -1),
                self.X_[mask],
            )

            X_pred, y_pred = X[~(mask)].reshape(X.shape[0], -1), X[mask]

            model = (
                self.tabpfn_clf
                if column_idx in self.categorical_features
                else self.tabpfn_reg
            )
            model.fit(X_train, y_train)
            embs += [model.get_embeddings(X_pred, additional_y=None)]

        return torch.cat(embs, 1).reshape(embs[0].shape[0], -1)


def efficient_random_permutation(indices, n_permutations=10):
    perms = []
    n_iter = 0
    while len(perms) < n_permutations and n_iter < n_permutations * 10:
        perm = efficient_random_permutation_(indices)
        if perm not in perms:
            perms += [perm]
        n_iter += 1

    return perms


def efficient_random_permutation_(indices):
    """
    Generate a single random permutation from a very large space.

    :param n: The size of the permutation (number of elements)
    :return: A list representing a random permutation of numbers from 0 to n-1
    """
    # Create a list of numbers from 0 to n-1
    permutation = indices

    # Shuffle the list in-place
    for i in range(len(indices) - 1, 0, -1):
        # Pick a random index from 0 to i
        j = random.randint(0, i)
        # Swap elements at i and j
        permutation[i], permutation[j] = permutation[j], permutation[i]

    return tuple(permutation)



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("your_package")
except PackageNotFoundError:
    __version__ = "0.1.0.dev0"

from .utils import is_tabpfn
from .utils import TabPFNRegressor, TabPFNClassifier

from . import utils_todo
